##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Bokeh plots.
"""
# stdlib
import warnings

# third-party
from IPython.core.display import HTML
from IPython.core.display import display as idisplay
from bokeh.embed import file_html
from bokeh.resources import CDN
from bokeh.models import LabelSet

# package
from idaes.vis.plotbase import PlotBase
from idaes.vis.plotutils import *


class BokehPlot(PlotBase):
    def __init__(self, current_plot=None):
        """Set the current plot in this constructor.

        Args:
            current_plot: A bokeh plot object.

        """
        active_dev_warning = '''The visualization library is still in active development and we hope
        to improve on it in future releases. Please use its functionality at your own discretion.'''
        warnings.warn(active_dev_warning, stacklevel=3)
        super().__init__(current_plot)
        self.html = ''
        self.filename = ''

    def annotate(self, x, y, label):
        """Annotate a plot with a given point and a label.

        Args:
            x: Value of independent variable.
            y: Value of dependent variable.
            label: Text label.

        Returns:
            None

        Raises:
            None
        """
        source = ColumnDataSource(data=dict(x=[x], y=[y], labels=[label]))
        self.current_plot.circle(x, y, line_color='red', fill_color='red')
        labels = LabelSet(
            x='x',
            y='y',
            text='labels',
            level='glyph',
            source=source,
            render_mode='canvas',
            text_font_size='9pt',
        )
        self.current_plot.add_layout(labels)

    def resize(self, height=-1, width=-1):
        """Resize a plot's height and width.

        Args:
            height: Height in screen units.
            width: Width in screen units.

        Returns:
            None

        Raises:
            None
        """
        if not self.current_plot:
            return
        if height != -1:
            self.current_plot.plot_height = height
        if width != -1:
            self.current_plot.plot_width = width

    def save(self, destination):
        """Save the current plot object to HTML in filepath provided by destination.

        Args:
            destination: Valid file path to save HTML to.

        Returns:
            filename where HTML is saved.

        Raises:
            None
        """
        if not self.html:
            self.html = file_html(self.current_plot, CDN, None)
        with open(destination, 'w') as output_file:
            output_file.write(self.html)
            self.filename = destination
        return self.filename

    def show(self, in_notebook=True):
        """Display plot in a Jupyter notebook.

        Args:
            self: Plot object.
            in_notebook: Display in Jupyter notebook or generate HTML file.

        Returns:
            None

        Raises:
            None
        """
        if in_notebook:
            self.html = file_html(self.current_plot, CDN, None)
            idisplay(HTML(self.html))
        else:
            show(self.current_plot)


class HeatExchangerNetwork(BokehPlot):
    def __init__(
        self,
        exchangers,
        stream_list,
        mark_temperatures_with_tooltips=False,
        mark_modules_with_tooltips=False,
        stage_width=2,
        y_stream_step=1,
    ):
        """Plot a heat exchanger network diagram.

        Args:
            exchangers: List of exchangers where each exchanger is a dict of the form:

                        .. code-block:: python

                            {'hot': 'H2', 'cold': 'C1', 'Q': 1400, 'A': 159, 'annual_cost': 28358,
                            'stg': 2}


                        where `hot` is the hot stream name, `cold` is the cold stream name, `A` is the area (in m^2), `annual_cost`
                        is the annual cost in $, `Q` is the amount of heat transferred from one stream
                        to another in a given exchanger and `stg` is the stage the exchanger belongs to.
                        The `utility_type`, if present, will specify if we plot the cold stream as water (:class:`idaes.vis.plot_utils.HENStreamType.cold_utility`)
                        or the hot stream as steam (:class:`idaes.vis.plot_utils.HENStreamType.hot_utility`).

                        Additionally, the exchanger could have the key `modules`, like this:

                        .. code-block:: python

                            {'hot': 'H1', 'cold': 'C1', 'Q': 667, 'A': 50, 'annual_cost': 10979, 'stg': 3,
                            'modules': {10: 1, 20: 2}}


                        The value of this key is a dictionary where each key is a module area and each value is how many
                        modules of that area are in the exchanger. It's indicated as a tooltip on the resulting diagram.

                        If a stream is involved in multiple exchanges in teh same stage, the stream will split into multiple
                        sub-streams with each sub-stream carrying one of the exchanges.

            stream_list: List of dicts representing streams where each item is a dict of the form:

                         .. code-block:: python

                            {'name':'H1', 'temps': [443, 435, 355, 333], 'type': HENStreamType.hot}


            mark_temperatures_with_tooltips: if True, we plot the stream temperatures and assign
                                             hover tooltips to them. Otherwise, we label them with text labels.

            mark_modules_with_tooltips: if True, we plot markers for modules (if present) and assign
                                             hover tooltips to them. Otherwise, we don't add module info.

            stage_width: How many units to use for each stage in the diagram (defaults to 2).

            y_stream_step: How many units to use to separate each stream/sub-stream from the next (defaults to 1).
        """

        total_streams_tuple = (-1, len(stream_list) + 1)
        total_cost = sum([exchanger['annual_cost'] for exchanger in exchangers])

        REGULAR_FONT_SIZE = "10pt"
        SMALL_FONT_SIZE = "5pt"
        COLD_STREAM_COLOR = "blue"
        HOT_STREAM_COLOR = "red"
        STREAM_TEMP_MARKER_FILL_COLOR = "white"
        STREAM_TEMP_MARKER_LINE_COLOR = HOT_STREAM_COLOR

        STAGE_WIDTH = stage_width
        STAGE_SEPARATOR = 0.5
        Y_STREAM_STEP = y_stream_step

        p = figure(
            x_range=total_streams_tuple,
            y_range=total_streams_tuple,
            title='Annualized Investment Cost = ${0}'.format(total_cost),
        )
        p.title.align = 'center'

        hot_streams = [
            {k['name']: k['temps']}
            for k in stream_list
            if k['type'] == HENStreamType.hot
        ]
        cold_streams = [
            {k['name']: k['temps']}
            for k in stream_list
            if k['type'] == HENStreamType.cold
        ]
        stage_set = set(sorted([exchanger['stg'] for exchanger in exchangers]))
        color_stage_dict = get_color_dictionary(stage_set)
        stream_y_values, hot_split_streams, cold_split_streams = get_stream_y_values(
            exchangers, hot_streams, cold_streams, y_stream_step=Y_STREAM_STEP
        )

        stage_x_limits = {}
        max_x = 0
        for stage in stage_set:
            if stage not in stage_x_limits:
                stage_x_limits[stage] = {
                    'x_start': max_x,
                    'x_exchanger_start': max_x + STAGE_SEPARATOR,
                    'x_exchanger_end': max_x + STAGE_WIDTH,
                    'x_true_end': max_x + STAGE_WIDTH + STAGE_SEPARATOR,
                }
                max_x = max_x + STAGE_WIDTH

        hot_stream_names = [list(stream.keys())[0] for stream in hot_streams]
        cold_stream_names = [list(stream.keys())[0] for stream in cold_streams]

        for i, stage in enumerate(stage_set):
            stage_exchangers = [
                exchanger
                for exchanger in exchangers
                if exchanger['stg'] == stage and not is_hot_or_cold_utility(exchanger)
            ]

            hot_split_streams_by_stage = [
                hot_split_stream[0]
                for hot_split_stream in hot_split_streams
                if hot_split_stream[2] == stage
            ]
            cold_split_streams_by_stage = [
                cold_split_stream[0]
                for cold_split_stream in cold_split_streams
                if cold_split_stream[2] == stage
            ]

            used_split_y_values_for_stage = []
            is_first_stage = i == 0
            is_last_stage = i == (len(stage_set) - 1)

            # Figure out distance between each exchanger for this stage
            # leaving some leeway for closing out split stream rectangles:
            start_x = stage_x_limits[stage]['x_exchanger_start']
            end_x = stage_x_limits[stage]['x_exchanger_end'] - STAGE_SEPARATOR
            if len(stage_exchangers) > 0:
                dist_between_exchanger = (end_x - start_x) / len(stage_exchangers)
            max_x = start_x

            for exchanger in stage_exchangers:
                exchanger_x_start = max_x + dist_between_exchanger
                hot_stream = exchanger['hot']
                cold_stream = exchanger['cold']
                if hot_stream in hot_split_streams_by_stage:
                    # Give me the 1st unused hot stream split y value for this stage:
                    for y_value in stream_y_values[hot_stream]['split_y_values']:
                        if y_value not in used_split_y_values_for_stage:
                            used_split_y_values_for_stage.append(y_value)
                            exchanger_y_start = y_value
                            break
                else:
                    # Just assign it to the default y value:
                    exchanger_y_start = stream_y_values[hot_stream]['default_y_value']

                if cold_stream in cold_split_streams_by_stage:
                    # Give me the 1st unused cold stream split y value for this stage:
                    for y_value in stream_y_values[cold_stream]['split_y_values']:
                        if y_value not in used_split_y_values_for_stage:
                            used_split_y_values_for_stage.append(y_value)
                            exchanger_y_end = y_value
                            break
                else:
                    # Just assign it to the default y value:
                    exchanger_y_end = stream_y_values[cold_stream]['default_y_value']

                p = plot_line_segment(
                    p,
                    exchanger_x_start,
                    exchanger_x_start,
                    exchanger_y_start,
                    exchanger_y_end,
                    color=color_stage_dict[stage],
                    legend="stage {0}".format(str(stage)),
                )
                p = add_exchanger_labels(
                    p,
                    exchanger_x_start,
                    exchanger_y_start,
                    exchanger_y_end,
                    SMALL_FONT_SIZE,
                    exchanger,
                    STREAM_TEMP_MARKER_LINE_COLOR,
                    STREAM_TEMP_MARKER_FILL_COLOR,
                    mark_modules_with_tooltips,
                )
                max_x += dist_between_exchanger
            for stream in stream_y_values.keys():
                color = (
                    HOT_STREAM_COLOR
                    if stream in hot_stream_names
                    else COLD_STREAM_COLOR
                )
                split_streams_list = (
                    hot_split_streams
                    if stream in hot_stream_names
                    else cold_split_streams
                )
                if (
                    stream not in hot_split_streams_by_stage
                    and stream not in cold_split_streams_by_stage
                ):
                    # Default y stream:
                    p = plot_line_segment(
                        p,
                        stage_x_limits[stage]['x_start'],
                        stage_x_limits[stage]['x_true_end'],
                        stream_y_values[stream]['default_y_value'],
                        stream_y_values[stream]['default_y_value'],
                        color=color,
                    )
                else:
                    # Split y stream:
                    split_count = [
                        stream_record[3]
                        for stream_record in split_streams_list
                        if stream_record[2] == stage and stream_record[0] == stream
                    ][0]
                    split_y_values = stream_y_values[stream]['split_y_values'][
                        0:split_count
                    ]
                    vertical_closer_y_start = max(
                        stream_y_values[stream]['default_y_value'], max(split_y_values)
                    )
                    vertical_closer_y_end = min(
                        stream_y_values[stream]['default_y_value'], min(split_y_values)
                    )
                    p = plot_line_segment(
                        p,
                        stage_x_limits[stage]['x_exchanger_start'],
                        stage_x_limits[stage]['x_exchanger_start'],
                        vertical_closer_y_start,
                        vertical_closer_y_end,
                        color=color,
                    )

                    p = plot_line_segment(
                        p,
                        stage_x_limits[stage]['x_exchanger_end'],
                        stage_x_limits[stage]['x_exchanger_end'],
                        vertical_closer_y_start,
                        vertical_closer_y_end,
                        color=color,
                    )
                    for y in sorted(split_y_values):
                        p = plot_line_segment(
                            p,
                            stage_x_limits[stage]['x_exchanger_start'],
                            stage_x_limits[stage]['x_exchanger_end'],
                            y,
                            y,
                            color=color,
                        )
            # 1st stage:
            # For hot streams:
            # - Plot line segments, stream names and highest tempreature label.
            # For cold streams:
            # - Plot arrows and lowest temperature label.
            # Plot hot utility exchanges.
            if is_first_stage:
                for stream in stream_y_values.keys():
                    temp_list = [
                        stream_dict['temps']
                        for stream_dict in stream_list
                        if stream_dict['name'] == stream
                    ][0]
                    if stream in hot_stream_names:
                        # Plot line segments that start/end a hot/cold stream
                        # and add relevant labels in the 1st stage:
                        p = plot_line_segment(
                            p,
                            stage_x_limits[stage]['x_start'],
                            stage_x_limits[stage]['x_exchanger_start'],
                            stream_y_values[stream]['default_y_value'],
                            stream_y_values[stream]['default_y_value'],
                            color=HOT_STREAM_COLOR,
                        )

                        temp_start_label = Label(
                            x=stage_x_limits[stage]['x_start'],
                            y=stream_y_values[stream]['default_y_value'],
                            x_offset=-10,
                            y_offset=2,
                            text_font_size=REGULAR_FONT_SIZE,
                            text=str(temp_list[0]),
                        )
                        p.add_layout(temp_start_label)

                        stream_name_label = Label(
                            x=stage_x_limits[stage]['x_start'],
                            y=stream_y_values[stream]['default_y_value'],
                            x_offset=-10,
                            y_offset=-12,
                            text_font_size=REGULAR_FONT_SIZE,
                            text=stream,
                        )
                        p.add_layout(stream_name_label)

                    if stream in cold_stream_names:
                        p = plot_stream_arrow(
                            p,
                            COLD_STREAM_COLOR,
                            temp_list[-1],
                            REGULAR_FONT_SIZE,
                            stage_x_limits[stage]['x_exchanger_start'],
                            stage_x_limits[stage]['x_start'],
                            stream_y_values[stream]['default_y_value'],
                            stream_y_values[stream]['default_y_value'],
                        )
                # plot steam exchange:
                steam_exchanger = [
                    exchanger
                    for exchanger in exchangers
                    if (
                        'utility_type' in exchanger
                        and exchanger['utility_type'] == HENStreamType.hot_utility
                    )
                ]
                if len(steam_exchanger) > 0:
                    cold_stream_name = steam_exchanger[0]['cold']
                    y_steam = stream_y_values[cold_stream_name]['default_y_value']
                    p.add_layout(
                        Arrow(
                            line_color=color_stage_dict[stage],
                            end=OpenHead(
                                line_color=color_stage_dict[stage],
                                line_width=2,
                                size=10,
                            ),
                            x_start=0.25,
                            y_start=y_steam - 0.25,
                            x_end=0.25,
                            y_end=y_steam + 0.25,
                        )
                    )
                    p = plot_line_segment(
                        p,
                        0.25,
                        0.25,
                        y_steam - 0.25,
                        y_steam + 0.25,
                        color=color_stage_dict[stage],
                        legend="stage {0}".format(str(stage)),
                    )
                    label_s = Label(
                        x=0.25,
                        y=y_steam - 0.25,
                        x_offset=-10,
                        y_offset=-15,
                        text_font_size=SMALL_FONT_SIZE,
                        text=steam_exchanger[0]['hot'],
                    )
                    p.add_layout(label_s)
                    label_q = Label(
                        x=0.25,
                        y=y_steam + 0.25,
                        x_offset=-5,
                        text_font_size=SMALL_FONT_SIZE,
                        text='{0} kW'.format(steam_exchanger[0]['Q']),
                    )
                    p.add_layout(label_q)

            # Last stage:
            # For cold streams:
            # - Plot line segments, stream names and highest tempreature label.
            # For hot streams:
            # - Plot arrows and lowest temperature label.
            # Plot cold utility exchanges.
            elif is_last_stage:
                for stream in stream_y_values.keys():
                    temp_list = [
                        stream_dict['temps']
                        for stream_dict in stream_list
                        if stream_dict['name'] == stream
                    ][0]
                    if stream in cold_stream_names:
                        # Plot line segments that start/end a hot/cold stream
                        # and add relevant labels in the 1st stage:
                        p = plot_line_segment(
                            p,
                            stage_x_limits[stage]['x_exchanger_end'],
                            stage_x_limits[stage]['x_true_end'],
                            stream_y_values[stream]['default_y_value'],
                            stream_y_values[stream]['default_y_value'],
                            color=COLD_STREAM_COLOR,
                        )

                        temp_start_label = Label(
                            x=stage_x_limits[stage]['x_true_end'],
                            y=stream_y_values[stream]['default_y_value'],
                            x_offset=-10,
                            y_offset=2,
                            text_font_size=REGULAR_FONT_SIZE,
                            text=str(temp_list[0]),
                        )
                        p.add_layout(temp_start_label)

                        stream_name_label = Label(
                            x=stage_x_limits[stage]['x_true_end'],
                            y=stream_y_values[stream]['default_y_value'],
                            x_offset=-10,
                            y_offset=-12,
                            text_font_size=REGULAR_FONT_SIZE,
                            text=stream,
                        )
                        p.add_layout(stream_name_label)

                    if stream in hot_stream_names:
                        p = plot_stream_arrow(
                            p,
                            HOT_STREAM_COLOR,
                            temp_list[-1],
                            REGULAR_FONT_SIZE,
                            stage_x_limits[stage]['x_exchanger_end'],
                            stage_x_limits[stage]['x_true_end'],
                            stream_y_values[stream]['default_y_value'],
                            stream_y_values[stream]['default_y_value'],
                        )
                # plot water exchange:
                water_exchanger = [
                    exchanger
                    for exchanger in exchangers
                    if (
                        'utility_type' in exchanger
                        and exchanger['utility_type'] == HENStreamType.cold_utility
                    )
                ]
                if len(water_exchanger) > 0:
                    hot_stream_name = water_exchanger[0]['hot']
                    y_water = stream_y_values[hot_stream_name]['default_y_value']
                    x_water = (
                        stage_x_limits[stage]['x_exchanger_end']
                        + (
                            stage_x_limits[stage]['x_true_end']
                            - stage_x_limits[stage]['x_exchanger_end']
                        )
                        / 2
                    )
                    p.add_layout(
                        Arrow(
                            line_color=color_stage_dict[stage],
                            end=OpenHead(
                                line_color=color_stage_dict[stage],
                                line_width=2,
                                size=10,
                            ),
                            x_start=x_water,
                            y_start=y_water + 0.25,
                            x_end=x_water,
                            y_end=y_water - 0.25,
                        )
                    )
                    p = plot_line_segment(
                        p,
                        x_water,
                        x_water,
                        y_water + 0.25,
                        y_water - 0.25,
                        color=color_stage_dict[stage],
                        legend="stage {0}".format(str(stage)),
                    )
                    label_s = Label(
                        x=x_water,
                        y=y_water - 0.25,
                        x_offset=-10,
                        y_offset=-15,
                        text_font_size=SMALL_FONT_SIZE,
                        text=water_exchanger[0]['cold'],
                    )
                    p.add_layout(label_s)
                    label_q = Label(
                        x=x_water,
                        y=y_water + 0.25,
                        x_offset=-5,
                        text_font_size=SMALL_FONT_SIZE,
                        text='{0} kW'.format(water_exchanger[0]['Q']),
                    )
                    p.add_layout(label_q)
            # Otherwise, plot line segments after the exchangers to finish off a stage:
            else:
                for stream in stream_y_values.keys():
                    color = (
                        COLD_STREAM_COLOR
                        if stream in cold_stream_names
                        else HOT_STREAM_COLOR
                    )
                    p = plot_line_segment(
                        p,
                        stage_x_limits[stage]['x_exchanger_end'],
                        stage_x_limits[stage]['x_true_end'],
                        stream_y_values[stream]['default_y_value'],
                        stream_y_values[stream]['default_y_value'],
                        color=color,
                    )
            # Plot stage temperatures for each stream
            for stream in stream_y_values.keys():
                temp_list = [
                    stream_dict['temps']
                    for stream_dict in stream_list
                    if stream_dict['name'] == stream
                ][0]
                if len(temp_list) > 1:
                    temp_list = temp_list[1:-1]
                if stage <= len(temp_list):
                    temp_x = (
                        stage_x_limits[stage]['x_exchanger_end']
                        + (
                            stage_x_limits[stage]['x_true_end']
                            - stage_x_limits[stage]['x_exchanger_end']
                        )
                        / 2
                    )
                    temp_to_mark = temp_list[stage - 1]
                    marker_temp = CircleCross(
                        x=temp_x,
                        y=stream_y_values[stream]['default_y_value'],
                        line_color=STREAM_TEMP_MARKER_LINE_COLOR,
                        fill_color=STREAM_TEMP_MARKER_FILL_COLOR,
                        size=10,
                    )
                    if mark_temperatures_with_tooltips:
                        source = ColumnDataSource(
                            data=dict(
                                x=[temp_x],
                                y=[stream_y_values[stream]['default_y_value']],
                                temps=['{0}'.format(str(temp_to_mark))],
                            )
                        )
                        glyph_marker = p.add_glyph(
                            source_or_glyph=source, glyph=marker_temp
                        )
                        hover_temps = HoverTool(
                            renderers=[glyph_marker],
                            tooltips=[('Temperature', '@temps')],
                        )
                        p.add_tools(hover_temps)
                    else:
                        glyph_marker = p.add_glyph(marker_temp)
                        temp_label = Label(
                            x=temp_x,
                            y=stream_y_values[stream]['default_y_value'],
                            text_font_size=SMALL_FONT_SIZE,
                            text='{0}'.format(str(temp_to_mark)),
                        )
                        p.add_layout(temp_label)

        p = turn_off_grid_and_axes_ticks(p)
        super().__init__(current_plot=p)


class ProfilePlot(BokehPlot):
    def __init__(
        self,
        data_frame,
        x='',
        y=None,
        title='',
        xlab='',
        ylab='',
        y_axis_type='auto',
        legend=None,
    ):
        """
        A profile plot includes 2 dependent variables and a single
        independent variable. Based on the Jupyter notebook `here <https://github.com/IDAES/model_contrib/blob/master/examples/mea_simple/mea_example_nb_01.ipynb>`_.

        Args:
            data_frame: a data frame with keys contained in x and y.
            x: Key in data-frame to use as x-axis.
            y: Keys in data-frame to use as y-axis.
            title: Title for a plot.
            xlab: Label for x-axis.
            ylab: Label for y-axis.
            y_axis_type: Specify "log" to pass logarithmic scale.
            legend : List of strings matching y.

        Returns:
            Plot object on success.

        Raises:
            ValueError: Bad input parameters
        """
        # validate inputs
        y = y or []  # coerce to list
        valid, msg = self.validate(data_frame, x, y, legend=legend)
        if not valid:
            raise ValueError(f"Invalid parameters: {msg}")
        # output to static HTML file (commented out for now)
        # output_file("profile_plot.html")
        p = figure(
            tools="pan,box_zoom,reset,save",
            title=title,
            x_axis_label=xlab,
            y_axis_label=ylab,
            y_axis_type=y_axis_type,
        )
        # Plotting both y1 and y2 against x:
        p.line(
            data_frame[x],
            data_frame[y[0]],
            legend=legend[0],
            line_color="green",
        )
        p.line(
            data_frame[x], data_frame[y[1]], legend=legend[1], line_color="blue"
        )
        super().__init__(current_plot=p)


## Not implemented:
# class PropertyModelPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                        ylab='', y_axis_type='auto', legend=[]):
#         """Draw pressure/enthalpy plots for different levels of temperature.
#
#         Args:
#             data_frame: a data frame with keys contained in x and y.
#             x: Key in data-frame to plot on x-axis.
#             y: Keys in data-frame to plot on y-axis.
#             title: Title for a plot.
#             xlab: Label for x-axis.
#             ylab: Label for y-axis.
#             y_axis_type: Specify "log" to pass logarithmic scale.
#             legend : List of strings matching y.
#
#         Returns:
#             Plot object on success.
#
#         Raises:
#             MissingVariablesException: Dependent variable or their data
#                                         not passed.
#             BadDataFrameException: No data-frame was generated for
#                                     the model object.
#         """
#         pass
#
#
# class GoodnessOfFitPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                         ylab='', y_axis_type='auto', legend=[]):
#         """Draw y against predicted value (y^) and display (calculate?) value of R^2.
#
#         Args:
#             data_frame: a data frame with keys contained in x and y.
#             x: Key in data-frame to use as x-axis.
#             y: Keys in data-frame to plot on y-axis.
#             title: Title for a plot.
#             xlab: Label for x-axis.
#             ylab: Label for y-axis.
#             y_axis_type: Specify "log" to pass logarithmic scale.
#             legend : List of strings matching y.
#
#         Returns:
#             Plot object on success.
#         Raises:
#             MissingVariablesException: Dependent variable or their data
#                                         not passed.
#             BadDataFrameException: No data-frame was generated for
#                                     the model object.
#         """
#         pass
#
#
# class TradeoffPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                  ylab='', y_axis_type='auto', legend=[]):
#         """Draw some parameter varying and the result on the objective value.
#
#         Args:
#             data_frame: a data frame with keys contained in x and y.
#             x: Key in data-frame to use as x-axis.
#             y: Keys in data-frame to plot on y-axis.
#             title: Title for a plot.
#             xlab: Label for x-axis.
#             ylab: Label for y-axis.
#             y_axis_type: Specify "log" to pass logarithmic scale.
#             legend : List of strings matching y.
#
#         Returns:
#             Plot object on success.
#
#         Raises:
#             MissingVariablesException: Dependent variable or their data
#                                         not passed.
#             BadDataFrameException: No data-frame was generated for
#                                     the model object.
#         """
#         pass
#
#
# class SensitivityPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                     ylab='', y_axis_type='auto', legend=[]):
#         """Need more information.
#         """
#         pass
#
#
# class ResidualPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                  ylab='', y_axis_type='auto', legend=[]):
#         """Plot x, some continuous value (e.g: T, P), against Y (% residual value).
#         Is this %-value calculated from variables in the idaes_model_object?
#
#         Args:
#             data_frame: a data frame with keys contained in x and y.
#             x: Key in data-frame to use as x-axis.
#             y: Keys in data-frame to plot on y-axis.
#             title: Title for a plot.
#             xlab: Label for x-axis.
#             ylab: Label for y-axis.
#             y_axis_type: Specify "log" to pass logarithmic scale.
#             legend : List of strings matching y.
#
#         Returns:
#             Plot object on success.
#         Raises:
#             MissingVariablesException: Dependent variable or their data
#                                         not passed.
#             BadDataFrameException: No data-frame was generated for
#                                     the model object.
#         """
#         pass
#
#
# class IsobarPlot(BokehPlot):
#     def __init__(self, data_frame, x='', y=[], title='', xlab='',
#                ylab='', y_axis_type='auto', legend=[]):
#         """Need more information.
#         """
#         pass
#
#
# class StreamTablePlot(BokehPlot):
#     def __init__(self, data_frame, title=''):
#         """Display a table for all names in the idaes_model_object_names
#         indexing rows according to row_start and row_stop.
#
#         Args:
#             data_frame: a data frame with keys contained in x and y.
#             title: Title for a plot.
#
#         Returns:
#             Plot object on success.
#         Raises:
#             MissingVariablesException: Dependent variable or their data
#                                         not passed.
#             BadDataFrameException: No data-frame was generated for
#                                     the model object.
#         """
#         pass



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example property package for the VLE calucations for the methanol synthesis
problem from Turkay & Grossmann. The parameters and correlations are from the
paper.
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Param, NonNegativeReals, Set
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import declare_process_block_class, PhysicalParameterBlock

from .methanol_state_block_VLE import IdealStateBlock

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("PhysicalParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class.
    """

    # Config block for the _IdealStateBlock
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("valid_phase", ConfigValue(
        default=('Vap', 'Liq'),
        domain=In(['Liq', 'Vap', ('Vap', 'Liq'), ('Liq', 'Vap')]),
        description="Flag indicating the valid phase",
        doc="""Flag indicating the valid phase for a given set of
    conditions, and thus corresponding constraints  should be included,
    **default** - ('Vap', 'Liq').
    **Valid values:** {
    **'Liq'** - Liquid only,
    **'Vap'** - Vapor only,
    **('Vap', 'Liq')** - Vapor-liquid equilibrium,
    **('Liq', 'Vap')** - Vapor-liquid equilibrium,}"""))

    CONFIG.declare("Cp", ConfigValue(
        default=0.035,
        domain=float,
        description="Constant pressure heat capacity in MJ/(kgmol K)",
        doc="""Value for the constant pressure heat capacity,
        **default** = 0.035 MJ/(kgmol K)"""))

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PhysicalParameterData, self).build()

        self.state_block_class = IdealStateBlock

        # List of valid phases in property package
        # List of valid phases in property package
        if self.config.valid_phase == ('Liq', 'Vap') or \
                self.config.valid_phase == ('Vap', 'Liq'):
            self.phase_list = Set(initialize=['Liq', 'Vap'],
                                  ordered=True)
        elif self.config.valid_phase == 'Liq':
            self.phase_list = Set(initialize=['Liq'])
        else:
            self.phase_list = Set(initialize=['Vap'])

        self.component_list = Set(initialize=['CH4',
                                              'CO',
                                              'H2',
                                              'CH3OH'])

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list,
                           "Vap": self.component_list}

        self.phase_equilibrium_idx = Set(initialize=[1, 2, 3, 4])

        self.phase_equilibrium_list = \
            {1: ["CH4", ("Vap", "Liq")],
             2: ["CO", ("Vap", "Liq")],
             3: ["H2", ("Vap", "Liq")],
             4: ["CH3OH", ("Vap", "Liq")]}

        # Gas Constant
        self.gas_constant = Param(within=NonNegativeReals,
                                  mutable=False,
                                  default=0.008314,
                                  doc='Gas Constant [MJ/(kgmol.K)]')

        self.vapor_pressure_coeff = {('CH4', 'A'): 15.2243,
                                     ('CH4', 'B'): 897.84,
                                     ('CH4', 'C'): -7.16,
                                     ('CO', 'A'): 14.3686,
                                     ('CO', 'B'): 530.22,
                                     ('CO', 'C'): -13.15,
                                     ('H2', 'A'): 13.6333,
                                     ('H2', 'B'): 164.9,
                                     ('H2', 'C'): 3.19,
                                     ('CH3OH', 'A'): 18.5875,
                                     ('CH3OH', 'B'): 3626.55,
                                     ('CH3OH', 'C'): -34.29}

        Cp = self.config.Cp
        Cv = Cp - self.gas_constant.value
        gamma = Cp / Cv

        self.gamma = Param(within=NonNegativeReals, mutable=True, default=gamma, doc='Ratio of Cp to Cv')

        self.Cp = Param(within=NonNegativeReals, mutable=True, default=Cp, doc='Constant pressure heat capacity [MJ/(kgmol K)]')

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {'flow_mol': {'method': None, 'units': 'kgmol/s'},
             'mole_frac': {'method': None, 'units': 'no unit'},
             'temperature': {'method': None, 'units': '100K'},
             'pressure': {'method': None, 'units': 'MPa'},
             'flow_mol_phase': {'method': None, 'units': 'kgmol/s'},
             'density_mol': {'method': '_density_mol',
                             'units': 'kgmol/m^3'},
             'vapor_pressure': {'method': '_vapor_pressure', 'units': 'MPa'},
             'mole_frac_phase': {'method': '_mole_frac_phase',
                                 'units': 'no unit'},
             'enthalpy_comp_liq': {'method': '_enthalpy_comp_liq',
                                   'units': 'MJ/kgmol'},
             'enthalpy_comp_vap': {'method': '_enthalpy_comp_vap',
                                   'units': 'MJ/kgmol'},
             'enthalpy_liq': {'method': '_enthalpy_liq',
                              'units': 'MJ/kgmol'},
             'enthalpy_vap': {'method': '_enthalpy_vap',
                              'units': 'MJ/kgmol'}})

        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'kg',
                               'amount': 'kgmol',
                               'temperature': '100K',
                               'energy': 'MJ',
                               'holdup': 'kgmol'})



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Make sure the supercritical steam cycle example solves.
"""

__author__ = "John Eslick"

import pytest
import pyomo.environ as pyo
from idaes.examples.power_generation.supercritical_steam_cycle import main
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              activated_equalities_generator)
from idaes.property_models import iapws95


solver_available = pyo.SolverFactory('ipopt').available()
prop_available = iapws95.iapws95_available()


@pytest.fixture(scope="module")
def initialize_model():
    return main()


def gross_power_mw(model):
    # pyo.value(m.fs.turb.power[0]) is the power consumed in Watts
    return -pyo.value(model.fs.turb.power[0])/1e6

@pytest.mark.slow
@pytest.mark.solver
@pytest.mark.skipif(not prop_available, reason="IAPWS not available")
@pytest.mark.skipif(not solver_available, reason="Solver not available")
def test_init(initialize_model):
    m, solver = initialize_model
    # check that the model solved properly and has 0 degrees of freedom
    assert(degrees_of_freedom(m)==0)
    for c in activated_equalities_generator(m):
        assert(abs(c.body() - c.lower) < 1e-4)


@pytest.mark.slow
@pytest.mark.solver
@pytest.mark.skipif(not prop_available, reason="IAPWS not available")
@pytest.mark.skipif(not solver_available, reason="Solver not available")
def test_init_value(initialize_model):
    m, solver = initialize_model
    assert gross_power_mw(m) == pytest.approx(620.8100259113626, abs=1e-3)


@pytest.mark.slow
@pytest.mark.solver
@pytest.mark.skipif(not prop_available, reason="IAPWS not available")
@pytest.mark.skipif(not solver_available, reason="Solver not available")
def test_valve_change(initialize_model):
    m, solver = initialize_model
    m.fs.turb.throttle_valve[1].valve_opening[:].value = 0.25
    solver.solve(m, tee=True)
    assert gross_power_mw(m) == pytest.approx(580.9566851114142, abs=1e-3)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example property package for the saponification of Ethyl Acetate with NaOH
Assumes dilute solutions with properties of H2O.
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           exp,
                           Param,
                           PositiveReals,
                           Set,
                           value,
                           Var)

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        ReactionParameterBlock,
                        ReactionBlockDataBase,
                        ReactionBlockBase)
from idaes.core.util.misc import add_object_reference

# Some more inforation about this module
__author__ = "Andrew Lee"


# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("SaponificationReactionParameterBlock")
class ReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """
    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(ReactionParameterData, self).build()

        self.reaction_block_class = ReactionBlock

        # List of valid phases in property package
        self.phase_list = Set(initialize=['Liq'])

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['H2O', 'NaOH',
                                              'EthylAcetate',
                                              'SodiumAcetate',
                                              'Ethanol'])

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {("R1", "Liq", "NaOH"): -1,
                                            ("R1", "Liq", "EthylAcetate"): -1,
                                            ("R1", "Liq", "SodiumAcetate"): 1,
                                            ("R1", "Liq", "Ethanol"): 1,
                                            ("R1", "Liq", "H2O"): 0}

        # Arrhenius Constant
        self.arrhenius = Param(default=3.132e6,
                               doc="Arrhenius constant")

        # Activation Energy
        self.energy_activation = Param(default=43000,
                                       doc="Activation energy [J/mol]")

        # Heat of Reaction
        dh_rxn_dict = {"R1": -49000}
        self.dh_rxn = Param(self.rate_reaction_idx,
                            initialize=dh_rxn_dict,
                            doc="Heat of reaction [J/mol]")

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({
                'k_rxn': {'method': '_rate_constant', 'units': 'm^3/mol.s'},
                'reaction_rate': {'method': "_rxn_rate", 'units': 'mol/m^3.s'}
                })
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


class _ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """
    def initialize(blk, outlvl=0, **kwargs):
        '''
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

                     * 0 = no output (default)
                     * 1 = report after each step

        Returns:
            None
        '''
        if outlvl > 0:
            _log.info('{} Initialization Complete.'.format(blk.name))


@declare_process_block_class("ReactionBlock",
                             block_class=_ReactionBlock)
class ReactionBlockData(ReactionBlockDataBase):
    """
    An example reaction package for saponification of ethyl acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(ReactionBlockData, self).build()

        # Create references to state vars
        # Concentration
        add_object_reference(self,
                             "conc_mol_comp_ref",
                             self.state_ref.conc_mol_comp)

        # Temperature
        add_object_reference(self,
                             "temperature_ref",
                             self.state_ref.temperature)

        # Heat of reaction - no _ref as this is the actual property
        add_object_reference(
                self,
                "dh_rxn",
                self.config.parameters.dh_rxn)

    # Rate constant method
    def _rate_constant(self):
        self.k_rxn = Var(initialize=1,
                         doc="Rate constant [m^3/mol.s]")

        try:
            self.arrhenius_eqn = Constraint(
                    expr=self.k_rxn == self._params.arrhenius *
                    exp(-self._params.energy_activation /
                        (self._params.gas_const*self.temperature_ref)))
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.k_rxn)
            self.del_component(self.arrhenius_eqn)
            raise

    # Rate of reaction method
    def _rxn_rate(self):
        self.reaction_rate = Var(self._params.rate_reaction_idx,
                                 initialize=0,
                                 doc="Rate of reaction [mol/m^3.s]")

        try:
            def rate_rule(b, r):
                return b.reaction_rate[r] == (
                            b.k_rxn*b.conc_mol_comp_ref["EthylAcetate"] *
                            b.conc_mol_comp_ref["NaOH"])
            self.rate_expression = Constraint(self._params.rate_reaction_idx,
                                              rule=rate_rule)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.reaction_rate)
            self.del_component(self.rate_expression)
            raise

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar

    def model_check(blk):
        pass



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Utility code for testing IDAES code.
"""
import os
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor


__author__ = "Dan Gunter"


def run_notebook(path: str, name: str):
    """Run a specific jupyter notebook 'name' located at `path`.
    """
    fullpath = os.path.join(path, name)
    with open(fullpath) as f:
        nb = nbformat.read(f, as_version=nbformat.NO_CONVERT)

    ep = ExecutePreprocessor(timeout=100)
    ep.allow_errors = True
    ep.preprocess(nb, {"metadata": {"path": path}})

    failed = False
    for cell in nb.cells:
        if "outputs" in cell:
            for output in cell["outputs"]:
                if output.output_type == "error":
                    failed = True
                    num = cell["execution_count"]
                    print(f"ERROR in {output['ename']} in {fullpath} [{num}]:")
                    for tb_line in output["traceback"]:
                        print(tb_line)
    return not failed



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Basis functions for generating the multiparameter equation of state
"""
import numpy as np
import sympy as sy


drd_vals = []
ar_vals = []
ai_vals = []
itt_val = 0
rtt_vals = []
dtrdt_vals = []
d2rd_vals = []
d3rd_vals = []
d4rd_vals = []
d5rd_Vals = []
critT, critD, critP, acc, R, M, Rm = [0, 0, 0, 0, 0, 0, 0]
molecule = ""
coeffs = []
indexes = []


def molData(fluidData, Dmolecule, RVal):
    """ Passing of the Data from the main module ::module:: MPEOSDeveloperModule

        :param fluidData: (critT, critP, critD, M, triple, acc).
        :type fluidData: array
        :param Dmolecule: Name of the molecule of interest.
        :type Dmolecule: str.
        :param RVal: Gas Constant.
        :type RVal: int.
    """

    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    molecule = Dmolecule
    (critT, critP, critD, M, triple, acc) = fluidData
    R = RVal
    Rm = RVal


def formCustomBasis(LemJac=False):
    """ Basis Functions developed a bank of terms based on literature (Lemmon, Span, Wagner) """
    global coeffs, indexes
    coeffs = []
    if LemJac:
        for i in range(1, 9):
            for j in range(1, 13):  # 12
                coeffs.append([i, j / 8.0, 0, 0])
        indexes.append(len(coeffs))
        for i in range(1, 6):
            for j in range(1, 24):  # 24
                coeffs.append([i, j / 8.0, 1, 0])
        for i in range(1, 6):
            for j in range(1, 30):  # 30
                coeffs.append([i, j / 8.0, 2, 0])
        for i in range(2, 5):
            for j in range(24, 38):  # 38
                coeffs.append([i, j / 2.0, 3, 0])
        indexes.append(len(coeffs))
        for i in range(1, 6):
            for j in range(1, 24):  # 24
                for m in range(1, 7):
                    coeffs.append([i, j / 8.0, 1, m / 2])
        for i in range(1, 6):
            for j in range(1, 30):  # 24
                for m in range(1, 7):
                    coeffs.append([i, j / 8.0, 2, m / 2])
        for i in range(2, 5):
            for j in range(24, 38):  # 38
                for m in range(1, 7):
                    coeffs.append([i, j / 2.0, 3, m / 2])
        indexes.append(len(coeffs))
    else:
        for i in range(1, 9):
            for j in range(1, 13):  # 12
                coeffs.append([i, j / 8.0, 0, 0])
        indexes.append(len(coeffs))
        for i in range(1, 6):
            for j in range(1, 24):  # 24
                coeffs.append([i, j / 8.0, 1, 0])
        for i in range(1, 6):
            for j in range(1, 30):  # 24
                coeffs.append([i, j / 8.0, 2, 0])
        for i in range(2, 5):
            for j in range(24, 38):  # 38
                coeffs.append([i, j / 2.0, 3, 0])
        indexes.append(len(coeffs))


def getTerm(Y):
    """ Prints index and basis function based on Y index"""
    for y in Y:
        print(y)
        d, t, c, m = coeffs[y - 1]
        print("Delta^%f Tau^%f np.exp(-Delta^%f) np.exp(-Tau^%f)" % (d, t, c, m))


def idealBY(D, T, Y, Beta):
    """Ideal Helmholtz contribution"""

    # TOLUENE
    if molecule == "TOL":
        a = [3.5241174832, 1.1360823464]
        c = [4.0, 0, 0]
        # B = [
        #     0.96464,
        #     -2.7855,
        #     0.86712,
        #     -0.18860,
        #     0.11804,
        #     0.0025181,
        #     0.57196,
        #     -0.029287,
        #     -0.43351,
        #     -0.12540,
        #     -0.028207,
        #     0.014076,
        # ]
        v = [1.6994, 8.0577, 17.059, 8.4567, 8.6423]
        u = [190, 797, 1619, 3072, 7915]

    # Carbon Monoxide
    if molecule == "CO":
        a = [-3.3728318564, 3.3683460039]
        c = [3.5, 0.22311e-6, 1.5]
        # B = [
        #     0.90554,
        #     -2.4515,
        #     0.53149,
        #     0.024173,
        #     0.072156,
        #     0.00018818,
        #     0.19405,
        #     -0.043268,
        #     -0.12778,
        #     -0.027896,
        #     -0.03414,
        #     0.016329,
        # ]
        v = [1.0128]
        u = [3089]

    # Water
    if molecule == "H2O":
        a = [-8.32044648201, 6.6832105268]
        c = [3.00632 + 1, 0, 0]
        v = [0.012436, 0.97315, 1.27950, 0.96956, 0.24873]
        u = [
            1.2878967 * float(critT),
            3.53734222 * float(critT),
            7.74073708 * float(critT),
            9.24437796 * float(critT),
            27.5075105 * float(critT),
        ]

    if c[2] > 0:
        idealA = (
            a[0]
            + a[1] * T
            + np.log(D)
            + (c[0] - 1) * np.log(T)
            - c[1] * (critT) ** c[2] / (c[2] * (c[2] + 1)) * T ** (-c[2])
        )
    else:
        idealA = a[0] + a[1] * T + np.log(D) + (c[0] - 1) * np.log(T)
    for uit, vi in zip(u, v):
        ui = uit / float(critT)
        idealA = idealA + vi * np.log(1 - np.exp(-ui * T))
    ai_vals = idealA
    return ai_vals


def arBY(D, T, Y, Beta):
    """Residual Helmholtz contribution"""
    global drd_vals
    global coeffs
    # drd_vals = [];
    val = 0
    # print len(Y)
    if type(Y) is not int:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                try:
                    val += y * (D ** d) * (T ** t)
                    pass
                except Exception:
                    print("Term", x, d, t, T, D)
            elif m == 0:
                val = np.exp(-D ** c) * (D ** (d)) * (T ** t)
            else:
                val = np.exp(-D ** c) * (D ** (d)) * (T ** t) * np.exp(-T ** m)

        return val


def drd(D, T):
    """Partial derivative with respect to density"""
    global drd_vals
    global coeffs

    drd_vals = []
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (D ** d) * (T ** t)
        elif m == 0:
            val = np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
        else:
            val = (
                np.exp(-D ** c) * (D ** (d)) * (T ** t) * np.exp(-T ** m) * (d - c * (D ** c))
            )
        drd_vals.append(val)


def d2rd(D, T):
    """Partial derivative with respect to density twice"""
    global d2rd_vals
    global coeffs
    d2rd_vals = []
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (d - 1) * (D ** d) * (T ** t)
        elif m == 0:  
            val = np.exp(-D ** c) * (
                (D ** (d))
                * (T ** t)
                * ((d - c * (D ** c)) * (d - 1 - c * (D ** c)) - (c ** 2) * (D ** c))
            )
        else:
            val = np.exp(-D ** c) * (
                (D ** (d))
                * (T ** t)
                * np.exp(-T ** m)
                * ((d - c * (D ** c)) * (d - 1 - c * (D ** c)) - (c ** 2) * (D ** c))
            )
        d2rd_vals.append(val)


def d2rdt(D, T):
    """Partial derivative with respect to density twice and temperature"""
    global d2rdt_vals
    global coeffs

    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (d - 1) * t * (D ** d) * (T ** t)
        elif m == 0: 
            val = np.exp(-D ** c) * (
                (D ** (d))
                * (T ** t)
                * ((d - c * (D ** c)) * (d - 1 - c * (D ** c)) - (c ** 2) * (D ** c))
            )

        else:
            val = (
                np.exp(-D ** c)
                * (
                    (D ** (d))
                    * (T ** t)
                    * np.exp(-T ** m)
                    * (
                        (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                        - (c ** 2) * (D ** c)
                    )
                )
                * (t - m * (T ** m))
            )
        d2rdt_vals.append(val)


def d3rd(D, T):
    """Third partial derivative with respect to density"""
    global d3rd_vals
    global coeffs
    d3rd_vals = []
    # i = 1
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (d - 1) * (d - 2) * (D ** d) * (T ** t)
        elif m == 0:
            # val = (D**(d))*(T**t)*exp(-(D**c))*(d*(d-1)*(d-2) + (D**c)*(-2*c + 6*d*c-3*(d**2)*c - 3*d*(c**2) + 3*(c**2) - (c**3)) + (D**(2*c))*(3*d*(c**2) -3*(c**2) + 3*(c**3)) - (c**3)*(D**(3*c)))
            val = (
                (D ** (d))
                * (T ** (t))
                * np.exp(-(D ** c))
                * (
                    (d - c * (D ** c)) * (d - 1 - c * (D ** c)) * (d - 2 - c * (D ** c))
                    + (3 * d - 3 + c - 3 * c * (D ** c)) * (-(c ** 2) * (D ** (c)))
                )
            )

        else:
            val = (
                (D ** (d))
                * (T ** t)
                * np.exp(-(T ** m))
                * np.exp(-(D ** c))
                * (
                    d * (d - 1) * (d - 2)
                    + (D ** c)
                    * (
                        -2 * c
                        + 6 * d * c
                        - 3 * (d ** 2) * c
                        - 3 * d * (c ** 2)
                        + 3 * (c ** 2)
                        - (c ** 3)
                    )
                    + (D ** (2 * c)) * (3 * d * (c ** 2) - 3 * (c ** 2) + 3 * (c ** 3))
                    - (c ** 3) * (D ** (3 * c))
                )
            )
        d3rd_vals.append(val)


def d4rd(D, T):
    """Fourth partial derivative with respect to density"""
    global d4rd_vals
    global coeffs
    d4rd_vals = []
    # i = 1
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (d - 1) * (d - 2) * (d - 3) * (D ** d) * (T ** t)
        elif m == 0:
            # val = (D**(d))*(T**t)*exp(-(D**c))*(d*(d-1)*(d-2) + (D**c)*(-2*c + 6*d*c-3*(d**2)*c - 3*d*(c**2) + 3*(c**2) - (c**3)) + (D**(2*c))*(3*d*(c**2) -3*(c**2) + 3*(c**3)) - (c**3)*(D**(3*c)))
            val = (
                (D ** (d))
                * (T ** (t))
                * np.exp(-(D ** c))
                * (
                    (d - c * (D ** c))
                    * (d - 1 - c * (D ** c))
                    * (d - 2 - c * (D ** c))
                    * (d - 3 - c * (D ** c))
                    + (
                        6 * (c ** 2) * D ** (2 * c)
                        - 7 * (c ** 2) * (D ** c)
                        + (c ** 2)
                        - 12 * c * d * (D ** c)
                        + 4 * c * d
                        + 18 * c * (D ** c)
                        - 6 * c
                        + 6 * d ** 2
                        - 18 * d
                        + 11
                    )
                    * (-(c ** 2) * (D ** (c)))
                )
            )

        else:  
            val = (
                (D ** (d))
                * (T ** t)
                * np.exp(-(T ** m))
                * np.exp(-(D ** c))
                * (
                    d * (d - 1) * (d - 2)
                    + (D ** c)
                    * (
                        -2 * c
                        + 6 * d * c
                        - 3 * (d ** 2) * c
                        - 3 * d * (c ** 2)
                        + 3 * (c ** 2)
                        - (c ** 3)
                    )
                    + (D ** (2 * c)) * (3 * d * (c ** 2) - 3 * (c ** 2) + 3 * (c ** 3))
                    - (c ** 3) * (D ** (3 * c))
                )
            )
            val = 0
        d4rd_vals.append(val)


def d5rd(D, T):
    """Fifth partial derivative with respect to density"""
    global d5rd_vals
    global coeffs
    d5rd_vals = []
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * (d - 1) * (d - 2) * (d - 3) * (d - 4) * (D ** d) * (T ** t)
        elif m == 0:
            x, y = sy.symbols("x, y")
            expr = x ** d * y ** t * sy.exp(-(x ** c))
            f_prime = expr.diff(x, 5)
            val = f_prime.subs([(x, D), (y, T)]) * (D ** 5)
        else: 
            val = (
                (D ** (d))
                * (T ** t)
                * np.exp(-(T ** m))
                * np.exp(-(D ** c))
                * (
                    d * (d - 1) * (d - 2)
                    + (D ** c)
                    * (
                        -2 * c
                        + 6 * d * c
                        - 3 * (d ** 2) * c
                        - 3 * d * (c ** 2)
                        + 3 * (c ** 2)
                        - (c ** 3)
                    )
                    + (D ** (2 * c)) * (3 * d * (c ** 2) - 3 * (c ** 2) + 3 * (c ** 3))
                    - (c ** 3) * (D ** (3 * c))
                )
            )
            val = 0
        d5rd_vals.append(val)


def dtrdt(D, T):
    """Second partial derivative with respect to density and temperature"""
    global dtrdt_vals
    global coeffs
    dtrdt_vals = []
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = d * t * (D ** d) * (T ** t)
        elif m == 0:
            val = np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
        else:
            val = (
                np.exp(-D ** c)
                * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
                * np.exp(-T ** m)
                * (t - m * (T ** m))
            )
        dtrdt_vals.append(val)


def rTT(D, T):
    """Second partial derivative with respect to temperature""" 
    global rtt_vals
    global coeffs
    rtt_vals = []
    for d, t, c, m in coeffs:
        val = 0
        if c == 0:
            val = t * (t - 1) * (D ** d) * (T ** (t))
        elif m == 0:
            val = np.exp(-D ** c) * ((D ** (d)) * (T ** (t))) * t * (t - 1)
        else:  # NOT DONE
            val = np.exp(-D ** c) * (D ** (d)) * (T ** t) * np.exp(-T ** t)
        rtt_vals.append(val)


def d3rdRes(D, T, Y, Beta):
    """Third partial derivative with respect to density"""
    global d3rd_vals
    global coeffs
    d3rd_vals = []
    val = 0
    for x, y in zip(Y, Beta):
        x = x - 1
        [d, t, c, m] = coeffs[x]
        if c == 0:
            val += y * d * (d - 1) * (d - 2) * (D ** d) * (T ** t)
        elif m == 0:  # TODO
            val += (
                y
                * np.exp(-D ** c)
                * (D ** (d))
                * (T ** t)
                * (
                    d * (d - 1) * (d - 2)
                    + c
                    * (D ** c)
                    * (-2 + 6 * d - 3 * (d ** 2) - 3 * d * c + 3 * c - (c ** 2))
                    + (D ** (2 * c)) * (3 * d * (c ** 2) - 3 * (c ** 2) + 3 * (c ** 3))
                    - (c ** 3) * (D ** (3 * c))
                )
            )
        else:
            val += (
                y
                * np.exp(-D ** c)
                * (D ** (d))
                * (T ** t)
                * np.exp(-T ** m)
                * (
                    d * (d - 1) * (d - 2)
                    + c
                    * (D ** c)
                    * (-2 + 6 * d - 3 * (d ** 2) - 3 * d * c + 3 * c - (c ** 2))
                    + (D ** (2 * c)) * (3 * d * (c ** 2) - 3 * (c ** 2) + 3 * (c ** 3))
                    - (c ** 3) * (D ** (3 * c))
                )
            )
        # print val
    return val


# PVT derivatives
def drdRes(D, T, Y, Beta):
    """
    Calculates the partial derivaties w.r.t. density
    Inputs:
        D - Delta
        T - Tau
        Y - index of basis Function (int or array)
        Beta - weighting (float or array)
    """
    global drd_vals
    global coeffs
    val = 0
    if type(Y) is not int:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                try:
                    val += y * d * (D ** d) * (T ** t)
                    pass
                except Exception:
                    print("Term", x, d, t, T, D)
            elif m == 0:
                val += y * np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
            else:
                val += (
                    y
                    * np.exp(-D ** c)
                    * (D ** (d))
                    * (T ** t)
                    * np.exp(-T ** m)
                    * (d - c * (D ** c))
                )
        return val
    else:
        x = Y - 1
        y = Beta
        [d, t, c, m] = coeffs[x]
        if c == 0:
            val += y * d * (D ** d) * (T ** t)
        elif m == 0:
            val += y * np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
        else:
            val += (
                y
                * np.exp(-D ** c)
                * (D ** (d))
                * (T ** t)
                * np.exp(-T ** m)
                * (d - c * (D ** c))
            )
        return val

# CV derivatives
def iTT(D, T):
    """Ideal helmholtz contribution second partial derivative with respect to temperature"""
    global itt_val

    # TOLUENE
    if molecule == "TOL":
        # a = [3.5241174832, 1.1360823464]
        c = [4.0, 0, 0]
        # B = [.96464, -2.7855, 0.86712, -0.18860, 0.11804, 0.0025181, 0.57196, -0.029287, -0.43351, -0.12540, -0.028207, 0.014076];
        v = [1.6994, 8.0577, 17.059, 8.4567, 8.6423]
        u = [190, 797, 1619, 3072, 7915]

    # Carbon Monoxide
    if molecule == "CO":
        # a = [-3.3728318564, 3.3683460039]
        c = [3.5, 0.22311e-6, 1.5]
        # B = [0.90554, -2.4515, 0.53149, 0.024173, 0.072156, 0.00018818, 0.19405, -0.043268, -0.12778, -0.027896, -0.03414, 0.016329]
        v = [1.0128]
        u = [3089]

    # Carbon Dioxide
    if molecule == "CO2":
        # a = [8.37304456, -3.70454304, 2.5]
        c = [2.5 + 1, 0, 0]
        v = [1.99427042, 0.62105248, 0.41195293, 1.04028922, 0.08327678]
        u = [
            3.15163 * float(critT),
            6.11190 * float(critT),
            6.77708 * float(critT),
            11.32384 * float(critT),
            27.0 * float(critT),
        ]

    # Water
    if molecule == "H2O":
        # a = [-8.32044648201, 6.6832105268]
        c = [3.00632 + 1, 0, 0]
        v = [0.012436, 0.97315, 1.27950, 0.96956, 0.24873]
        u = [
            1.2878967 * float(critT),
            3.53734222 * float(critT),
            7.74073708 * float(critT),
            9.24437796 * float(critT),
            27.5075105 * float(critT),
        ]

    itt_val = 0
    if c[2] != 0:
        itt_val = -(c[0] - 1) - c[1] * (critT ** c[2]) / (c[2] * (c[2] + 1)) * (
            T ** (-c[2])
        ) * (-c[2]) * (-c[2] - 1)
        # /T**2;
    else:
        itt_val = -(c[0] - 1)
    for uit, vi in zip(u, v):
        ui = uit / float(critT)
        itt_val = itt_val - vi * ((ui) ** 2) * (T ** 2) * np.exp(-ui * T) * (
            (1 - np.exp(-ui * T)) ** (-2)
        )
    return itt_val


def rTTRes(D, T, Y, Beta):
    """Residual helmholtz contribution second partial derivative with respect to temperature"""
    global rtt_vals
    global coeffs
    val = 0
    if type(Y) is int:
        x = Y - 1
        y = Beta
        [d, t, c, m] = coeffs[x]
        if c == 0:
            value = t * (t - 1) * (D ** d) * (T ** (t))
        elif m == 0:  
            value = np.exp(-D ** c) * ((D ** (d)) * (T ** (t))) * t * (t - 1)
        else:
            value = (
                np.exp(-D ** c)
                * ((D ** (d)) * (T ** (t)))
                * ((t - m * (T ** m)) * (t - 1 - m * (T ** m)) - (m ** 2) * (T ** m))
            )
        val = y * value
    else:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                value = t * (t - 1) * (D ** d) * (T ** (t))
            elif m == 0: 
                value = np.exp(-D ** c) * ((D ** (d)) * (T ** (t))) * t * (t - 1)
            else:
                value = (
                    np.exp(-D ** c)
                    * ((D ** (d)) * (T ** (t)))
                    * (
                        (t - m * (T ** m)) * (t - 1 - m * (T ** m))
                        - (m ** 2) * (T ** m)
                    )
                )
            val += y * value
    return val


def d2rdRes(D, T, Y, Beta):
    """Residual helmholtz contribution second partial derivative with respect to density"""
    global d2rd_vals
    global coeffs
    val = 0
    d2rd_vals = []
    if type(Y) is int:
        x = Y - 1
        y = Beta
        [d, t, c, m] = coeffs[x]
        if c == 0:
            val = y * d * (d - 1) * (D ** d) * (T ** t)
        elif m != 0:  # TODO
            val = np.exp(-D ** c) * (D ** (d)) * (T ** t) * np.exp(-T ** t)
        else:
            val = (
                y
                * np.exp(-D ** c)
                * (
                    (D ** (d))
                    * (T ** t)
                    * (
                        (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                        - (c ** 2) * (D ** c)
                    )
                )
            )
    else:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                val += y * d * (d - 1) * (D ** d) * (T ** t)
            elif m == 0:  # TODO
                val += (
                    y
                    * np.exp(-D ** c)
                    * (
                        (D ** (d))
                        * (T ** t)
                        * (
                            (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                            - (c ** 2) * (D ** c)
                        )
                    )
                )
            else:
                val += (
                    y
                    * np.exp(-D ** c)
                    * (
                        (D ** (d))
                        * (T ** t)
                        * np.exp(-T ** m)
                        * (
                            (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                            - (c ** 2) * (D ** c)
                        )
                    )
                )
    return val


def dtrdtRes(D, T, Y, Beta):
    """Residual helmholtz contribution second partial derivative with respect to density and temperature"""
    global dtrdt_vals
    global coeffs
    val = 0
    dtrdt_vals = []
    if type(Y) is int:
        x = Y - 1
        y = Beta
        [d, t, c, m] = coeffs[x]
        if c == 0:
            val = y * d * t * (D ** d) * (T ** t)
        elif m == 0:  # TODO
            val = y * t * np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
        else:
            val = (
                y
                * t
                * np.exp(-D ** c)
                * ((D ** (d)) * (T ** t) * np.exp(-T ** m) * (d - c * (D ** c)))
                * (t - m * (T ** m))
            )
    else:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                val += y * d * t * (D ** d) * (T ** t)
            elif m == 0:
                val += (
                    y * t * np.exp(-D ** c) * ((D ** (d)) * (T ** t) * (d - c * (D ** c)))
                )
            else:
                val += (
                    y
                    * t
                    * np.exp(-D ** c)
                    * ((D ** (d)) * (T ** t) * np.exp(-T ** m) * (d - c * (D ** c)))
                    * (t - m * (T ** m))
                )
    return val


def d2rdrtRes(D, T, Y, Beta):
    """Residual helmholtz contribution third partial derivative with respect to density(2) and temperature(1)"""
    global d2rd_vals
    global coeffs
    val = 0
    d2rd_vals = []
    if type(Y) is int:
        x = Y - 1
        y = Beta
        [d, t, c, m] = coeffs[x]
        if c == 0:
            val = y * d * (d - 1) * (D ** d) * t * (T ** t)
        elif m != 0:
            val = (
                np.exp(-D ** c)
                * (D ** (d))
                * (T ** t)
                * np.exp(-T ** t)
                * (t - m * (T ** m))
            )
        else:
            val = (
                y
                * np.exp(-D ** c)
                * (
                    (D ** (d))
                    * (T ** t)
                    * (
                        (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                        - (c ** 2) * (D ** c)
                    )
                )
                * t
            )
    else:
        for x, y in zip(Y, Beta):
            x = x - 1
            [d, t, c, m] = coeffs[x]
            if c == 0:
                val += y * d * (d - 1) * (D ** d) * t * (T ** t)
            elif m == 0:
                val += (
                    y
                    * np.exp(-D ** c)
                    * (
                        (D ** (d))
                        * (T ** t)
                        * (
                            (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                            - (c ** 2) * (D ** c)
                        )
                    )
                    * t
                )
            else:
                val += (
                    y
                    * np.exp(-D ** c)
                    * (
                        (D ** (d))
                        * (T ** t)
                        * np.exp(-T ** m)
                        * (
                            (d - c * (D ** c)) * (d - 1 - c * (D ** c))
                            - (c ** 2) * (D ** c)
                        )
                    )
                    * (t - m * (T ** m))
                )
    return val
    



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Command Line Interface for IDAES DMF.

Uses "Click" to handle command-line parsing and dispatch.
"""
# stdlib
from collections import namedtuple
import datetime
from enum import Enum
import json
import logging
from operator import itemgetter
import pathlib
import sys
from typing import List
from urllib.parse import urlparse, ParseResult
import yaml

# third-party
import click
import humanize
import pendulum

# package
from idaes.dmf import DMF, DMFConfig, resource
from idaes.dmf import errors
from idaes.dmf.workspace import Fields
from idaes.dmf import util
from idaes.dmf.util import ColorTerm, yaml_load

__author__ = "Dan Gunter"

_log = logging.getLogger(__name__)


class Code(Enum):
    """Return codes from the CLI.
    """

    OK = 0
    WORKSPACE_NOT_FOUND = 1
    CONFIGURATION_NOT_FOUND = 2
    NOT_SUPPORTED = 3
    IMPORT_RESOURCE = 4
    DMF = 5
    DMF_OPER = 6
    INPUT_VALUE = 7
    CANCELED = 8


# Some common fields we can show
SHOW_FIELDS = "type, desc, creator, created, modified, fikes, codes, version"


def level_from_verbosity(vb):
    level = 0  # for pycharm
    if vb >= 3:
        level = logging.DEBUG
    elif vb == 2:
        level = logging.INFO
    elif vb == 1:
        level = logging.WARN
    elif vb == 0:
        level = logging.ERROR
    elif vb == -1:
        level = logging.FATAL
    elif vb <= -2:
        level = logging.FATAL + 1
    return level


class AliasedGroup(click.Group):
    """Improved click.Group that will accept unique prefixes for the
    commands, as well as a set of aliases.

    For example, the following code will create `mycommand` as a group,
    and alias the subcommand "info" to invoke the subcommand "status".
    Any unique prefix of "info" (not conflicting with other subcommands or
    aliases) or "status" will work, e.g. "inf" or "stat"::

        @click.group(cls=AliasedGroup, aliases={"info": "status"})
        def mycommand():
            pass
    """

    def __init__(self, aliases=None, **attrs):
        super().__init__(**attrs)
        self._aliases = aliases

    def get_command(self, ctx, cmd_name):
        command = click.Group.get_command(self, ctx, cmd_name)
        if command is None:
            commands = self.list_commands(ctx)
            # if `cmd_name` is an alias, map to canonical name
            matches = [key for key in self._aliases if key.startswith(cmd_name)]
            if len(matches) == 1:
                # check that this alias does not also match a command directly
                alias_matches = [x for x in commands if x.startswith(cmd_name)]
                if len(alias_matches) > 0:
                    ctx.fail(
                        f"Alias prefix {cmd_name}->{matches[0]} also matches command"
                        f"{'s' if len(alias_matches) > 1 else ''}"
                        f": {' '.join(sorted(alias_matches))}"
                    )
                # rename to canonical (matched) command
                cmd_name = self._aliases[matches[0]]
            elif len(matches) > 1:
                ctx.fail(
                    f"Prefix '{cmd_name}' matches multiple aliases: "
                    f"{' '.join(sorted(matches))}"
                )
            # try to match prefix of command
            matches = [x for x in commands if x.startswith(cmd_name)]
            if len(matches) == 1:
                command = click.Group.get_command(self, ctx, matches[0])
            elif len(matches) > 1:
                ctx.fail(
                    f"Command '{cmd_name}' could be a prefix for multiple commands: "
                    f"{' '.join(sorted(matches))}"
                )
        # Return found command
        return command  # matched, or None


class URLType(click.ParamType):
    """Click type for URLs.
    """

    name = "URL"
    envvar_list_splitter = ","

    def convert(self, value, param, ctx):
        if isinstance(value, ParseResult):
            result = value
        else:
            result = urlparse(value)
        return result


@click.group(
    cls=AliasedGroup,
    aliases={
        "describe": "status",
        "add": "register",
        "resource": "info",
        "show": "info",
        "list": "ls",
        "delete": "rm",
        "graph": "related",
    },
    help="Data management framework command wrapper",
)
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Increase verbosity. Show warnings if given once, "
    "then info, and then debugging messages.",
)
@click.option(
    "--quiet",
    "-q",
    count=True,
    help="Increase quietness. If given once, "
    "only show critical messages. If "
    "given twice, show no messages.",
)
def base_command(verbose, quiet):
    """Data management framework command wrapper.

    This command does nothing by itself except provide global
    options and list subcommands.
    """
    if quiet > 0 and verbose > 0:
        raise click.BadArgumentUsage("Options for verbosity and quietness conflict")
    if verbose > 0:
        _log.setLevel(level_from_verbosity(verbose))
    else:
        _log.setLevel(level_from_verbosity(-quiet))


@click.command(
    help="Initialize the current workspace. Optionally, create a new workspace."
)
@click.argument("path", type=click.Path())
@click.option(
    "--create/--no-create",
    default=False,
    help="Create new workspace. If `--name` and `--desc` are not provided, these will"
    " be prompted for interactively.",
)
@click.option("--name", type=click.STRING, help="Workspace name")
@click.option("--desc", type=click.STRING, help="Workspace description")
@click.option(
    "--html",
    type=click.Path(exists=True, resolve_path=True),
    default=None,
    help="Path to built HTML documentation",
)
def init(path, create, name, desc, html):
    """Initialize the current workspace used for the data management framework commands.
    Optionally, create a new workspace.
    """
    _log.info(f"Initialize with workspace path={path}")
    if create:
        _log.info("Create new workspace")
        # pre-check that there is no file/dir by this name
        try:
            if pathlib.Path(path).exists():
                click.echo(f"Cannot create workspace: path '{path}' already exists")
                sys.exit(Code.DMF_OPER.value)
        except PermissionError:
            click.echo(f"Cannot create workspace: path '{path}' not accessible")
            sys.exit(Code.DMF_OPER.value)
        if not name:
            name = click.prompt("New workspace name")
        if not desc:
            desc = click.prompt("New workspace description")
        if html is None:
            # guess html path
            # XXX: don't try to verify the guess
            errfile = pathlib.Path(errors.__file__)
            docsdir = errfile.parent.parent.parent / 'docs'
            hpath = [str(docsdir / 'build')]
        else:
            hpath = [html]
        try:
            d = DMF(path=path, create=True, name=name, desc=desc, html_paths=hpath)
        except errors.WorkspaceError as err:
            click.echo(f"Cannot create workspace: {err}")
            sys.exit(Code.DMF_OPER.value)
        click.echo(f"Configuration in '{d.configuration_file}")
    else:
        _log.info("Use existing workspace")
    # In either case, switch to provided config
    try:
        _ = DMF(path=path, create=False, save_path=True)  # noqa: F841
    except errors.WorkspaceConfNotFoundError:
        click.echo(f"Workspace configuration not found at path='{path}'")
        sys.exit(Code.WORKSPACE_NOT_FOUND.value)
    except errors.WorkspaceNotFoundError:
        click.echo(f"Existing workspace not found at path='{path}'")
        click.echo("Add --create flag to create a workspace.")
        sys.exit(Code.WORKSPACE_NOT_FOUND.value)


@click.command(help="Get status of workspace")
@click.option("--color/--no-color", default=True, help="Use color for output")
@click.option(
    "--show",
    "-s",
    type=click.Choice(["files", "htmldocs", "logging", "all"]),
    multiple=True,
)
@click.option(
    "--all", "-a", "show_all", flag_value="yes", help="Synonym for `--show all`"
)
def status(color, show, show_all):
    if show_all == "yes":
        show = ["all"]
    _log.debug(f"Get status. Show items: {' '.join(show) if show else '(basic)'}")
    t = ColorTerm(enabled=color)
    if not DMFConfig.configuration_exists():
        click.echo(f"No configuration found at '{DMFConfig.configuration_path()}'")
        sys.exit(Code.CONFIGURATION_NOT_FOUND.value)
    try:
        d = DMF()
    except errors.WorkspaceConfNotFoundError as err:
        _log.fatal(f"Cannot get status: {err}")
        click.echo(str(err))
        sys.exit(Code.WORKSPACE_NOT_FOUND.value)

    # pretty-display a key/value pair or list value
    def item(key, value=None, before="", color=t.green):
        after_key = "" if key == "" else ":"
        if value is None:
            return f"{before}{color}{key}{after_key}{t.reset}"
        elif key is None:
            return f"{before}{color}{value}{t.reset}"
        return f"{before}{color}{key}{after_key}{t.reset} {value}"

    indent_spc = "  "

    print(item("settings", color=t.blue))
    indent = indent_spc
    conf = DMFConfig()  # note: must exist due to earlier check
    for key, value in conf.c.items():
        print(item(key, value, before=indent))

    print(item("workspace", color=t.blue))
    indent = indent_spc
    for key, value in (
        ("location", d.root),
        ("name", d.name),
        ("description", d.description),
        ("created", d.meta[d.CONF_CREATED]),
        ("modified", d.meta[d.CONF_MODIFIED]),
    ):
        print(item(key, value, before=indent))

    _show_optional_workspace_items(d, show, indent_spc, item, t=t)


def _show_optional_workspace_items(d, items, indent_spc, item_fn, t=None):
    indent = indent_spc
    for thing in items:
        if thing == "files" or thing == "all":
            print(item_fn("files", before=indent))
            fpath = pathlib.Path(d.datafiles_path)
            fdirs = [dr for dr in fpath.glob("*") if dr.is_dir()]
            indent = indent_spc * 2
            print(item_fn("count", len(fdirs), before=indent))
            total_size = sum(
                (sum((fp.stat().st_size for fp in fd.glob("*"))) for fd in fdirs)
            )
            print(
                item_fn("total_size", humanize.naturalsize(total_size), before=indent)
            )
        if thing == "htmldocs" or thing == "all":
            indent = indent_spc
            doc_paths = d.get_doc_paths()
            print(item_fn("html_documentation_paths", before=indent))
            indent = indent_spc * 2
            for p in doc_paths:
                print(item_fn("-", p, before=indent))
        if thing == "logging" or thing == "all":
            indent = indent_spc
            print(item_fn("logging", before=indent))
            log_conf = d.meta.get(Fields.LOG_CONF, None)
            indent = 2 * indent_spc
            if log_conf is None:
                print(item_fn(None, "not configured", before=indent, color=t.yellow))
            else:
                for subconf in sorted(log_conf.keys()):
                    print(item_fn(subconf, before=indent))
                    indent2 = indent + indent_spc
                    for i2, v2 in log_conf[subconf].items():
                        print(item_fn(i2, v2, before=indent2))


@click.command(help="Register a new object in the DMF workspace")
@click.argument("url", type=URLType(), metavar="FILE")
@click.option("--info", help="Show info on created resource", flag_value="yes")
@click.option(
    "--copy/--no-copy",
    default=True,
    help="Whether input file is copied into DMF "
    "workspace, or referred to by location",
)
@click.option(
    "--type",
    "-t",
    "resource_type",
    type=click.Choice(tuple(resource.RESOURCE_TYPES)),
    help="Resource type (default=determined from file)",
)
@click.option(
    "--strict/--no-strict",
    help="If inferring the type fails, "
    "with `--strict` report an error, and with `--no-strict` fall back "
    "to importing as a generic file",
)
@click.option(
    "--unique/--no-unique",
    default=True,
    help="Stop if another resource has a file matching this file's "
    "name and contents",
)
@click.option(
    "--contained",
    help="Add a 'contained in' relation to the given resource",
    metavar="RESOURCE-ID",
    multiple=True,
)
@click.option(
    "--derived", help="Add 'derived from' relation", metavar="OBJECT-ID", multiple=True
)
@click.option(
    "--used",
    help="Add 'used by' relation to the given resource",
    metavar="RESOURCE-ID",
    multiple=True,
)
@click.option(
    "--prev",
    help="Add 'version of previous' relation to the given resource",
    metavar="RESOURCE-ID",
    multiple=True,
)
@click.option(
    "--is-subject",
    flag_value="yes",
    help="If given, this resource will be the subject rather than the object "
    "of any and all relations added.",
)
@click.option(
    "--version",
    help="Set semantic version for this resource (default=0.0.0)",
    default=None,
)
def register(
    resource_type,
    url,
    info,
    copy,
    strict,
    unique,
    contained,
    derived,
    used,
    prev,
    is_subject,
    version,
):
    _log.debug(f"Register object type='{resource_type}' url/path='{url.path}'")
    # process url
    if url.scheme in ("file", ""):
        path = url.path
    else:
        click.echo("Currently, URL must be a file")
        sys.exit(Code.NOT_SUPPORTED.value)
    # create the resource
    _log.debug("create resource")
    try:
        rsrc = resource.Resource.from_file(
            path, as_type=resource_type, strict=strict, do_copy=copy
        )
    except resource.Resource.InferResourceTypeError as err:
        click.echo(f"Failed to infer resource: {err}")
        sys.exit(Code.IMPORT_RESOURCE.value)
    except resource.Resource.LoadResourceError as err:
        click.echo(f"Failed to load resource: {err}")
        sys.exit(Code.IMPORT_RESOURCE.value)
    # connect to DMF
    try:
        dmf = DMF()
    except errors.WorkspaceError as err:
        click.echo(f"Failed to connect to DMF: {err}")
        sys.exit(Code.WORKSPACE_NOT_FOUND.value)
    except errors.DMFError as err:
        click.echo(f"Failed to connect to DMF: {err}")
        sys.exit(Code.DMF.value)
    # check uniqueness
    if unique:
        df = rsrc.v["datafiles"][0]  # file info for this upload
        query = {"datafiles": [{"sha1": df["sha1"]}]}
        query_result, dup_ids = dmf.find(query), []
        for dup in query_result:
            dup_df = dup.v["datafiles"][0]
            if dup_df["path"] in df["path"]:
                dup_ids.append(dup.id)
        n_dup = len(dup_ids)
        if n_dup > 0:
            click.echo(
                f"This file is already in {n_dup} resource(s): " f"{' '.join(dup_ids)}"
            )
            sys.exit(Code.DMF_OPER.value)
    # process relations
    _log.debug("add relations")
    rel_to_add = {  # translate into standard relation names
        resource.PR_CONTAINS: contained,
        resource.PR_DERIVED: derived,
        resource.PR_USES: used,
        resource.PR_VERSION: prev,
    }
    target_resources = {}  # keep target resources in dict, update at end
    for rel_name, rel_ids in rel_to_add.items():
        for rel_id in rel_ids:
            if rel_id in target_resources:
                rel_subj = target_resources[rel_id]
            else:
                rel_subj = dmf.fetch_one(rel_id)
                target_resources[rel_id] = rel_subj
            if rel_subj is None:
                click.echo(f"Relation {rel_name} target not found: {rel_id}")
                sys.exit(Code.DMF_OPER.value)
            if is_subject == "yes":
                resource.create_relation_args(rsrc, rel_name, rel_subj)
            else:
                resource.create_relation_args(rel_subj, rel_name, rsrc)
            _log.debug(f"added relation {rsrc.id} <-- {rel_name} -- {rel_id}")
    _log.debug("update resource relations")
    for rel_rsrc in target_resources.values():
        dmf.update(rel_rsrc)
    # add metadata
    if version:
        try:
            vlist = resource.version_list(version)
        except ValueError:
            click.echo(f"Invalid version `{version}`")
            sys.exit(Code.INPUT_VALUE.value)
        else:
            rsrc.v["version_info"]["version"] = vlist
    # add the resource
    _log.debug("add resource begin")
    try:
        new_id = dmf.add(rsrc)
    except errors.DuplicateResourceError as err:
        click.echo(f"Failed to add resource: {err}")
        sys.exit(Code.DMF_OPER.value)
    _log.debug(f"added resource: {new_id}")
    if info == "yes":
        pfxlen = len(new_id)
        si = _ShowInfo("term", pfxlen)
        for rsrc in dmf.find_by_id(new_id):
            si.show(rsrc)
    else:
        click.echo(new_id)


@click.command(help="List resources in the workspace")
@click.option("--color/--no-color", default=True, help="Use color for output")
@click.option(
    "--show",
    "-s",
    multiple=True,
    help=f"Show given field(s) in listing. Common fields: {SHOW_FIELDS}",
)
@click.option(
    "--sort",
    "-S",
    "sort_by",
    type=click.Choice(["id", "type", "desc", "created", "modified", "version", "name"]),
    multiple=True,
    help="Sort by given field; if repeated, combine to make a compound sort key",
)
@click.option(
    "--prefix/--no-prefix",
    "prefix",
    help="By default, shown 'id' is the shortest unique prefix; "
    "`--no-prefix` shows full id",
    default=True,
)
@click.option("--reverse", "-r", "reverse", flag_value="yes", help="Reverse sort order")
def ls(color, show, sort_by, reverse, prefix):
    try:
        d = DMF()
    except errors.WorkspaceError as e:
        print(f"Workspace error: {e}")
        sys.exit(Code.WORKSPACE_NOT_FOUND.value)
    if not show:
        show = ["type", "desc", "modified"]  # note: 'id' is always first
    else:
        try:
            show = _split_and_validate_fields(show)
        except ValueError as err:
            click.echo(f"Bad fields for --show option: {err}")
            sys.exit(Code.INPUT_VALUE.value)
    reverse = bool(reverse == "yes")
    if not sort_by:
        sort_by = ["id"]
    resources = list(d.find())
    _print_resource_table(resources, show, sort_by, reverse, prefix, color)


def _split_and_validate_fields(fields: List[str]) -> List[str]:
    """Split comma-separated 'show' fields, validate that they are
    part of a resource, and return a list of tuples.

    Raises:
        ValueError: if the fields do not validate
    Returns:

    """
    dummy = resource.DUMMY_RESOURCE
    # Find every field in the resource
    result = []
    for grp in fields:
        for f in grp.split(","):
            f = f.strip()
            if f in _show_fields:
                result.append(f)
                continue
            if "." in f:
                keys, v = f.split("."), dummy
                try:
                    for k in keys:
                        try:
                            v = v[k]
                        except TypeError:
                            raise ValueError(f"bad field {f}")
                except KeyError:
                    raise ValueError(f"bad field {f} (error at " f"sub-field {k})")
                result.append(tuple(keys))
            elif f not in dummy:
                raise ValueError(f"bad field {f}")
            else:
                result.append(f)
    return result


def _print_resource_table(resources, show_fields, sort_by, reverse, prefix, color):
    """Text-mode `ls`.
    """
    t = ColorTerm(enabled=color)
    if len(resources) == 0:
        print("no resources to display")
        return
    uuid_len = util.uuid_prefix_len([r.id for r in resources])
    full_len = max((len(r.id) for r in resources)) if not prefix else uuid_len
    fields = ["id"] + list(show_fields)
    nfields = len(fields)
    # calculate table body. do this first to get widths.
    hdr_fields = [".".join(f) if isinstance(f, tuple) else f for f in fields]
    rows, maxwid, widths = [], 60, [len(f) for f in hdr_fields]
    for r in resources:
        row = []
        for i, fld in enumerate(fields):
            # transform field for display
            try:
                transformer = _show_fields[fld]
            except KeyError:
                transformer = _IdentityField(fld)
            transformer.set_value(r)
            # if it's a UUID field, add info about unique prefix length
            is_id_field = isinstance(transformer, _IdField)
            if is_id_field:
                transformer.term = t
                transformer.pfxlen = uuid_len
                transformer.flen = full_len
            # extract display string and set field width
            s = str(transformer)
            slen = full_len if is_id_field else len(s)
            if slen > widths[i]:
                if slen > maxwid:
                    s, widths[i] = s[:maxwid], maxwid
                else:
                    widths[i] = slen
            row.append(s)
        # append sort keys (not displayed)
        sort_obj = [_show_fields[fld] for fld in sort_by]
        for o in sort_obj:
            o.set_value(r)
        row.extend([o.value for o in sort_obj])
        # add row to table body
        rows.append(row)
    # sort rows
    nsort = len(sort_by)
    if nsort > 0:
        sort_key_idx = tuple(range(nfields, nfields + nsort))
        rows.sort(key=itemgetter(*sort_key_idx), reverse=reverse)
    # print table header
    hdr_columns = [t.bold + "{{f:{w}s}}".format(w=w).format(f=f)
                   for f, w in zip(hdr_fields, widths)]
    print(" ".join(hdr_columns) + t.normal)
    # print table body
    for row in rows:
        col = []
        for i, fld in enumerate(fields):
            if i == 0:
                col.append(t.red)
            elif fld == resource.Resource.TYPE_FIELD:
                col.append(t.yellow)
            elif fld != "desc":
                col.append(t.green)
            else:
                col.append("")
        row_columns = [
            f"{col[i]}{f:{w}}{t.reset}"
            for i, f, w in zip(range(len(widths)), row[:nfields], widths)
        ]
        print(" ".join(row_columns))


@click.command(help="Find resources in the workspace")
@click.option("--color/--no-color", default=True, help="Use color for output")
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["list", "info", "json"]),
    default="list",
    help="Output format",
)
@click.option(
    "--show",
    "-s",
    multiple=True,
    help=f"Show given field(s) in listing. Common fields: {SHOW_FIELDS}",
)
@click.option(
    "--sort",
    "-S",
    "sort_by",
    type=click.Choice(["id", "type", "desc", "created", "modified", "version"]),
    multiple=True,
    help="Sort by given field; if repeated, combine to make a compound sort key",
)
@click.option("--reverse", "-r", "reverse", flag_value="yes", help="Reverse sort order")
@click.option(
    "--prefix/--no-prefix",
    "prefix",
    help="By default, shown 'id' is the shortest unique prefix; "
    "`--no-prefix` shows full id",
    default=True,
)
@click.option("--by", default="", help="Creator name")
@click.option("--created", default="", help="Creation date or date range")
@click.option("--file", "filedesc", default="", help="File desc(ription)")
@click.option("--modified", default="", help="Modification date or date range")
@click.option("--name", default="", help="Matches any of the aliases")
@click.option("--type", "datatype", default="", help="The resource type")
def find(
    output_format,
    color,
    show,
    sort_by,
    prefix,
    reverse,
    by,
    created,
    filedesc,
    modified,
    name,
    datatype,
):
    d = DMF()
    if output_format == "list":
        if not show:
            show = ["type", "desc", "modified"]  # note: 'id' is always first
        else:
            try:
                show = _split_and_validate_fields(show)
            except ValueError as err:
                click.echo(f"Bad fields for --show option: {err}")
                sys.exit(Code.INPUT_VALUE.value)
    reverse = bool(reverse == "yes")
    if not sort_by:
        sort_by = ["id"]

    # Build query
    query = {}
    if by:
        query["creator.name"] = by
    if created:
        try:
            query["created"] = _date_query(created)
        except ValueError as err:
            click.echo(f"bad date for 'created': {err}")
            sys.exit(Code.INPUT_VALUE.value)
    if filedesc:
        query["datafiles"] = [{"desc": filedesc}]
    if modified:
        try:
            query["modified"] = _date_query(modified)
        except ValueError as err:
            click.echo(f"bad date for 'modified': {err}")
            sys.exit(Code.INPUT_VALUE.value)
    if name:
        query["aliases"] = [name]
    if datatype:
        query["type"] = datatype

    # Execute query
    _log.info(f"find: query = '{query}'")
    _log.debug("find.begin")
    resources = list(d.find(query))
    _log.debug("find.end")

    # Print result
    if output_format == "list":
        # print resources like `ls`
        _print_resource_table(resources, show, sort_by, reverse, prefix, color)
    elif output_format == "info":
        # print resources one by one
        si = _ShowInfo("term", 32, color=color)
        for rsrc in resources:
            si.show(rsrc)
    elif output_format == "json":
        # print resources as JSON
        for r in resources:
            print(json.dumps(r.v, indent=2))


def _date_query(datestr):
    """Transform date(s) into a query.
    """

    def _ts(x):  # get timestamp of date or time
        try:
            return x.timestamp()
        except AttributeError:
            return datetime.datetime(*x.timetuple()[:6]).timestamp()

    dates = datestr.split("..", 1)
    _log.debug(f"date '{datestr}', split: {dates}")
    try:
        parsed_dates = [
            pendulum.parser.parse(d, exact=True, tz=pendulum.local_timezone())
            if d
            else None
            for d in dates
        ]
    except pendulum.exceptions.ParserError as err:
        raise ValueError(str(err))
    if len(parsed_dates) == 2:
        begin_date, end_date = parsed_dates
        query = {}
        if begin_date:
            query['$ge'] = _ts(begin_date)
        if end_date:
            query['$le'] = _ts(end_date)
    else:
        pd = parsed_dates[0]
        if isinstance(pd, pendulum.Pendulum):
            _log.warning("datetime given, must match to the second")
            query = _ts(pd)
        else:
            query = {'$ge': _ts(pd), '$le': _ts(pd) + 60 * 60 * 24}  # 1 day
    return query


@click.command(help="Show detailed information about a resource")  # aliases: resource
@click.argument("identifier")
@click.option("--multiple/--no-multiple", default=False)
@click.option(
    "--format",
    "-f",
    "output_format",
    type=click.Choice(["term", "json", "jsonc"]),
    multiple=False,
    default="term",
    help="Output format: `term` (the default) shows terminal-friendly output, with "
    "some colors; `json` outputs JSON text; `jsonc` outputs compact JSON (no indents"
    "or line breaks)",
)
@click.option(
    "--color/--no-color",
    default=True,
    help="In term mode, use (or do not use)" " color for output",
)
def info(identifier, multiple, output_format, color):
    _log.debug(f"info for resource id='{identifier}'")
    try:
        resource.identifier_str(identifier, allow_prefix=True)
    except ValueError as err:
        click.echo(f"{err}")
        sys.exit(Code.INPUT_VALUE.value)
    rsrc_list = list(find_by_id(identifier))
    n = len(rsrc_list)
    if n > 1 and not multiple:
        click.echo(
            f"Too many ({n}) resources match prefix '{identifier}'. "
            "Add option --multiple to allow multiple matches."
        )
        sys.exit(Code.DMF_OPER.value)
    elif n == 0:
        click.echo("Resource not found")
        sys.exit(Code.DMF_OPER.value)
    pfxlen = len(identifier)
    si = _ShowInfo(output_format, pfxlen, color=color)
    for rsrc in rsrc_list:
        si.show(rsrc)


@click.command(help="Show resources related (connected) to a given resource")
@click.argument("identifier")
@click.option(
    "-d",
    "--direction",
    type=click.Choice(["out", "in"]),
    default="out",
    help="Direction of relationship to show (default=out[going])",
)
@click.option(
    "--color/--no-color",
    default=True,
    help="In term mode, use (or do not use)" " color for output",
)
@click.option(
    "--unicode/--no-unicode",
    default=True,
    help="With `--unicode`, allow unicode characters for connecting "
    "output items with 'lines'. With --no-unicode, use plain ASCII "
    "characters for this",
)
def related(identifier, direction, color, unicode):
    _log.info(f"related to resource id='{identifier}'")
    t = ColorTerm(enabled=color)
    dmf = DMF()
    try:
        resource.identifier_str(identifier, allow_prefix=True)
    except ValueError as err:
        click.echo(f"{err}")
        sys.exit(Code.INPUT_VALUE.value)
    _log.debug(f"begin: finding root resource {identifier}")
    rsrc_list = list(find_by_id(identifier, dmf=dmf))
    n = len(rsrc_list)
    if n > 1:
        click.echo(f"Too many resources matching `{identifier}`")
        sys.exit(Code.INPUT_VALUE)
    rsrc = rsrc_list[0]
    _log.debug(f"end: finding root resource {identifier}")
    # get related resources
    _log.debug(f"begin: finding related resources for {identifier}")
    outgoing = direction == "out"
    rr = list(dmf.find_related(rsrc, meta=["aliases", "type"], outgoing=outgoing))
    _log.debug(f"end: finding related resources for {identifier}")
    # stop if no relations
    if not rr:
        _log.warning(f"no resource related to {identifier}")
        click.echo(f"No relations for resource `{identifier}`")
        sys.exit(0)
    _log.info(f"got {len(rr)} related resources")
    # debugging
    if _log.isEnabledFor(logging.DEBUG):
        dbgtree = '\n'.join(['  ' + str(x) for x in rr])
        _log.debug(f"related resources:\n{dbgtree}")
    # extract uuids & determine common UUID prefix length
    uuids = [item[2][resource.Resource.ID_FIELD] for item in rr]
    pfx = util.uuid_prefix_len(uuids)
    # initialize queue with depth=1 items
    q = [item for item in rr if item[0] == 1]
    # print root resource
    print(_related_item(rsrc.id, rsrc.name, rsrc.type, pfx, t, unicode))
    # set up printing style
    if unicode:
        # connector chars
        vrt, vrd, relbow, relbow2, rarr = (
            '\u2502',
            '\u2506',
            '\u2514',
            '\u251C',
            '\u2500\u2500',
        )
        # relation prefix and arrow
        relpre, relarr = (
            ['\u2500\u25C0\u2500\u2524', '\u2524'][outgoing],
            ['\u2502', '\u251C\u2500\u25B6'][outgoing],
        )
    else:
        # connector chars
        vrt, vrd, relbow, relbow2, rarr = '|', '.', '+', '+', '--'
        # relation prefix and arrow
        relpre, relarr = ['<-[', '-['][outgoing], [']-', ']->'][outgoing]
    # create map of #items at each level, so we can easily
    # know when there are more at a given level, for drawing
    n_at_level = {0: 0}
    for item in rr:
        depth = item[0]
        if depth in n_at_level:
            n_at_level[depth] += 1
        else:
            n_at_level[depth] = 1
    # print tree
    while q:
        depth, rel, meta = q.pop()
        n_at_level[depth] -= 1
        indent = ''.join(
            [
                f" {t.blue}{vrd if n_at_level[i - 1] else ' '}{t.resetc} "
                for i in range(1, depth + 1)
            ]
        )
        print(f"{indent} {t.blue}{vrt}")
        rstr = f"{t.blue}{relpre}{t.yellow}{rel.predicate}"\
            f"{t.blue}{relarr}{t.reset}"
        if meta["aliases"]:
            item_name = meta["aliases"][0]
        else:
            item_name = meta.get("desc", "-")
        istr = _related_item(
            meta[resource.Resource.ID_FIELD], item_name, meta["type"], pfx, t, unicode
        )
        # determine correct connector (whether there is another one down the stack)
        elbow = relbow if (not q or q[-1][0] != depth) else relbow2
        print(f"{indent} {t.blue}{elbow}{rarr}{t.resetc}{rstr} {istr}")
        new_rr = []
        for d2, rel2, _ in rr:
            if outgoing:
                is_same = rel2.subject == rel.object
            else:
                is_same = rel2.object == rel.subject
            if d2 == depth + 1 and is_same:
                q.append((d2, rel2, _))
            else:
                new_rr.append((d2, rel2, _))
        rr = new_rr


def _related_item(id_, name, type_, pfx, t, unicode):
    return f"{t.red}{id_[:pfx]} {t.green}{type_} {t.resetc}{name}"


@click.command(help="Remove a resource")  # aliases: delete
@click.argument("identifier")
@click.option(
    "-y",
    "--yes",
    flag_value="yes",
    help="No interactive confirmations; assume 'yes' answer to all",
)
@click.option("--list/--no-list", "list_resources", default=True)
@click.option("--multiple/--no-multiple", default=False)
def rm(identifier, yes, multiple, list_resources):
    _log.info(f"remove resource '{identifier}'")
    try:
        resource.identifier_str(identifier, allow_prefix=True)
    except ValueError as errmsg:
        click.echo(f"Invalid identifier. Details: {errmsg}")
        sys.exit(Code.INPUT_VALUE.value)
    rsrc_list = list(find_by_id(identifier))
    found_multiple = len(rsrc_list) > 1
    if found_multiple and not multiple:
        click.echo(
            f"Too many ({len(rsrc_list)}) resources match prefix '{identifier}'. "
            "Add option --multiple to allow multiple matches."
        )
        sys.exit(Code.DMF_OPER.value)
    fields = ["type", "desc", "modified"]  # "id" is prepended by _ls_basic()
    if list_resources:
        _print_resource_table(rsrc_list, fields, ["id"], False, False, True)
    if yes != "yes":
        if found_multiple:
            s = f"these {len(rsrc_list)} resources"
        else:
            s = "this resource"
        do_remove = click.confirm(f"Remove {s}", prompt_suffix="? ", default=False)
        if not do_remove:
            click.echo("aborted")
            sys.exit(Code.CANCELED.value)
    d = DMF()
    for r in rsrc_list:
        _log.debug(f"begin remove-resource id={r.id}")
        d.remove(identifier=r.id)
        _log.debug(f"end remove-resource id={r.id}")
    if found_multiple:
        s = f"{len(rsrc_list)} resources removed"
    else:
        s = "resource removed"
    click.echo(s)


######################################################################################


def find_by_id(identifier, dmf=None):
    if dmf is None:
        dmf = DMF()
    return dmf.find_by_id(identifier)


class _ShowInfo:
    """Container for methods, etc. to show info about a resource.
    """

    contents_indent, json_indent = 4, 2  # for `term` output

    def __init__(self, output_format, pfxlen, color=None, unicode=True):
        self._terminal = ColorTerm(enabled=color)
        self._pfxlen = pfxlen
        self._fmt = output_format
        self._resource = None
        C = namedtuple('Corners', ['nw', 'ne', 'se', 'sw'])
        if unicode:
            self._corners = C._make('\u250C\u2510\u2518\u2514')
            self._hz, self._vt = '\u2500', '\u2502'
        else:
            self._corners = C._make('++++')
            self._hz, self._vt = '-', '|'

    def show(self, rsrc):
        self._resource = rsrc
        getattr(self, f"show_{self._fmt}")()

    def show_json(self):
        json.dump(self._resource.v, sys.stdout, indent=self.json_indent)
        print()

    def show_jsonc(self):
        json.dump(self._resource.v, sys.stdout)
        print()

    def show_term(self):
        t = self._terminal
        rval = self._human_readable_values()
        term_width = t.width
        width = min(term_width, self._longest_line(rval) + 3 + self.contents_indent)
        self._print_info_term_header(width)
        top_keys = sorted(rval.keys())
        for rownum, tk in enumerate(top_keys):
            val = rval[tk]
            if self._has_values(val):
                contents_str = yaml.dump(
                    yaml_load(json.dumps(val)),
                    default_flow_style=False,
                    explicit_end=False,
                )
                if contents_str.endswith('...\n'):
                    contents_str = contents_str[:-4]
                print(
                    f"{t.cyan}{self._vt}{t.resetc} {t.bold}{t.cyan}{tk}{t.reset}"
                    f"{' ' * (width - len(tk) - 3)}{t.cyan}{self._vt}{t.resetc}"
                )
                self._print_info_contents_term(contents_str, width)
        print(
            f"{t.cyan}{self._corners.sw}{self._hz * (width - 2)}"
            f"{self._corners.se}{t.resetc}"
        )

    def _longest_line(self, formatted_resource):
        longest = 0
        for k, v in formatted_resource.items():
            v_longest = max(
                (len(s) for s in json.dumps(v, indent=self.json_indent).split('\n'))
            )
            longest = max((longest, v_longest))
        return longest

    def _print_info_term_header(self, width):
        t, r = self._terminal, self._resource
        padding = width - len(r.id) - 10
        if padding % 2 == 1:
            lpad = padding // 2 - 1
            rpad = padding // 2 - 1
        else:
            lpad, rpad = padding // 2 - 1, padding // 2 - 2
        print(
            f"{t.cyan}{self._corners.nw}{lpad * self._hz}{t.resetc} Resource "
            f"{t.red}{r.id[:self._pfxlen]}"
            f"{t.green}{r.id[self._pfxlen:]} "
            f"{t.cyan}{rpad * self._hz}{self._corners.ne}{t.resetc}"
        )

    def _print_info_contents_term(self, s, width):
        t, n = self._terminal, self.contents_indent
        contents_width = width - 2 - n
        indent = " " * n
        for line in s.split('\n'):
            p = 0
            while p < len(line):
                print_line = line[p : p + contents_width]
                print(
                    f"{t.cyan}{self._vt}{t.resetc}{indent}{print_line}"
                    f"{' ' * (contents_width - len(print_line))}{t.cyan}{self._vt}"
                    f"{t.resetc}"
                )
                p += contents_width

    def _human_readable_values(self):
        val, result = self._resource.v, {}

        def dateize(v):
            return pendulum.fromtimestamp(v).to_datetime_string()

        for k, v in val.items():
            if k in ("created", "modified"):
                result[k] = dateize(v)
            elif k == "version_info":
                result[
                    "version"
                ] = f"{resource.format_version(v['version'])} @ {dateize(v['created'])}"
            elif k == "relations":
                relations = []
                for rel in v:
                    is_object = rel["role"] == "object"
                    predicate = f"--[{rel['predicate']}]--"
                    if is_object:
                        s = f"{rel['identifier']} {predicate}> ME"
                    else:
                        s = f"{rel['identifier']} <{predicate} ME"
                    relations.append(s)
                if relations:
                    result[k] = relations
            else:
                if isinstance(v, list):
                    result[k] = v[:]
                elif isinstance(v, dict):
                    result[k] = v.copy()
                else:
                    result[k] = v
        return result

    @staticmethod
    def _has_values(data):
        if isinstance(data, list) or isinstance(data, dict):
            return bool(data)
        return True


class _LsField:
    """Subclasses define how to retrieve and transform the
    user-specified field into a string for display by 'ls'.
    """

    def __init__(self, key):
        self._key = key
        self.value = ""

    def set_value(self, rsrc):
        if isinstance(self._key, list) or isinstance(self._key, tuple):
            v = rsrc.v
            for k in self._key:
                v = v[k]
            self.value = v
        else:
            self.value = rsrc.v[self._key]


class _IdentityField(_LsField):
    def __str__(self):
        return str(self.value)


class _DateField(_LsField):
    def __str__(self):
        return pendulum.fromtimestamp(self.value).to_datetime_string()


class _FilesField(_LsField):
    def __str__(self):
        if len(self.value) > 1:
            return f"{self.value[0]['path']} ..."
        elif len(self.value) == 1:
            return f"{self.value[0]['path']}"
        return "<none>"


class _CodesField(_LsField):
    def __str__(self):
        if len(self.value) > 1:
            return f"{self.value[0]['name']} ..."
        elif len(self.value) == 1:
            return f"{self.value[0]['name']}"
        return "<none>"


class _IdField(_LsField):
    def __init__(self, key):
        super().__init__(key)
        self.pfxlen = 0
        self.flen = 0
        self.term = None

    def __str__(self):
        t = self.term
        if self.pfxlen < len(self.value):
            if self.flen > self.pfxlen:
                return (
                    f"{t.cyan}{self.value[: self.pfxlen]}{t.resetc}"
                    f"{self.value[self.pfxlen: self.flen]}"
                )
            else:
                return self.value[: self.pfxlen]
        return self.value


class _VersionField(_LsField):
    def __str__(self):
        return resource.format_version(self.value)


class _NameField(_LsField):
    def __init__(self):
        super().__init__("aliases")
        self.value = []

    def __str__(self):
        if len(self.value) == 0:
            return ""
        return self.value[0]

# Mapping from the field name to an instance of a subclass
# of _Field that can extract sortable and formatted values.


_show_fields = {
    "id": _IdField(resource.Resource.ID_FIELD),
    "type": _IdentityField("type"),
    "desc": _IdentityField("desc"),
    "creator": _IdentityField(("creator", "name")),
    "created": _DateField("created"),
    "modified": _DateField("modified"),
    "files": _FilesField("datafiles"),
    "codes": _CodesField("codes"),
    "version": _VersionField(("version_info", "version")),
    "name": _NameField()
}


######################################################################################

# Register base commands
base_command.add_command(init)
base_command.add_command(register)
base_command.add_command(status)
base_command.add_command(ls)
base_command.add_command(find)
base_command.add_command(info)
base_command.add_command(related)
base_command.add_command(rm)

if __name__ == '__main__':
    base_command()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test utility functions
"""
# standard library
import importlib
import logging
import os
import shutil
import tempfile
# third party
import pytest
# local
from idaes.dmf import util, dmfbase

from mock import MagicMock, patch

scratchdir = None


def init_logging():
    """Init logging for tests.
    """
    log = logging.getLogger('idaes.dmf')
    h = logging.StreamHandler()
    f = logging.Formatter(
        fmt='%(asctime)s %(name)s [%(levelname)s] %(message)s')
    h.setFormatter(f)
    log.addHandler(h)
    if os.environ.get('TEST_DEBUG', ''):
        log.setLevel(logging.DEBUG)


def eq_(var, val, msg=None):
    """Shorthand for common test assertion.
    """
    if msg:
        assert var == val, msg
    else:
        assert var == val


def ne_(var, val, msg=None):
    """Shorthand for common test assertion.
    """
    if msg:
        assert var != val, msg
    else:
        assert var != val


def black_hole():
    """Output destination that swallows up everything.
    """
    return open('/dev/null', 'w')


def patch_modules():
    """Using mock.patch, modify sys.modules to mock up any
    modules not actually importable for tests.

    Returns:
        (mock.mock._patch_dict) Object returned by `mock.patch.dict()`.
    """
    sm = mock_import('core.process_base', 'idaes_models', assign={
        'core.process_base:ProcessBase':
            'idaes.dmf.tests.process_base:ProcessBase'
    })
    # .. can add more to `sm` dict here..
    if sm is not None:
        result = patch.dict('sys.modules', sm)
    else:
        result = None
    return result


def mock_import(name, package, assign=None):
    """Set up mocking for imports of "<package>.<name>".
    Optionally, assign real (importable) modules and/or module members
    to the mocked ones (see `assign`).

    Args:
        name (str): Module path to mock up
        package (str): Package name to mock up
        assign (dict): Mapping where key is mock module and
           value is real module path. Optionally, a ':<member>' can
           be appended to both key and value to assign members of the
           real module to the mock. If this is None, nothing is done.

    Returns:
        (dict) New values for sys.modules dict to be patched.
          If the given `package` imports without error,
          nothing is done and return value is None.
    """
    name, package = name.strip(), package.strip()
    assert name, 'module must be non-empty'
    assert package, 'package must be non-empty'
    mlist = name.split('.')
    try:
        importlib.import_module(package, mlist[0])
        return None
    except ImportError:
        pass
    pkg = MagicMock()
    # add module path to mock
    modpath, curmod = package, pkg
    sys_modules = {modpath: curmod}
    for m in mlist:
        modpath = modpath + '.' + m
        curmod = getattr(curmod, m)
        sys_modules[modpath] = curmod
    # assign real modules/classes to mock ones
    if assign:
        for mock_imp, real_imp in assign.items():
            # extract class, if present
            if ':' in mock_imp:
                mock_imp, mock_class = mock_imp.split(':')
                real_imp, real_class = real_imp.split(':')
            else:
                mock_class, real_class = None, None
            # import real module
            real_pkg = real_imp.split('.')[0]
            real_modules = '.' + real_imp[real_imp.index('.') + 1:]
            real_mod = importlib.import_module(real_modules, real_pkg)
            # assign real module, or class, to mock one
            mock_mod, mock_imp_list = pkg, mock_imp.split('.')
            if mock_class:
                # set mock.module.path.Class = real.module.path.Class
                for m in mock_imp_list:
                    mock_mod = getattr(mock_mod, m)
                real_classobj = getattr(real_mod, real_class)
                setattr(mock_mod, mock_class, real_classobj)
            else:
                # set mock.module.path = real.module.path
                for m in mock_imp_list[:-1]:
                    mock_mod = getattr(mock_mod, m)
                setattr(mock_mod, mock_imp_list[-1], real_mod)
    return sys_modules


class TempDir(object):
    def __init__(self, chdir=False):
        self._d = None
        self._chdir = chdir
        self._origdir = None

    def __enter__(self):
        self._d = tempfile.mkdtemp(suffix='-dmf')
        if self._chdir:
            self._origdir = os.getcwd()
            os.chdir(self._d)
        return self._d

    def __exit__(self, *args):
        if self._d is not None:
            # remove files in dir
            rmdirs = []
            for dirpath, subdirs, files in os.walk(self._d):
                for f in files:
                    os.unlink(os.path.join(dirpath, f))
                rmdirs.append(dirpath)
            # remove dirs
            while rmdirs:
                path = rmdirs.pop()
                os.rmdir(path)
        self._d = None
        if self._chdir:
            os.chdir(self._origdir)


@pytest.fixture(scope='function')
def tmp_dmf():
    """Test fixture to create a DMF in a temporary
    directory.
    """
    global scratchdir
    tmpdir = tempfile.mkdtemp()
    tmpdmf = dmfbase.DMF(path=tmpdir, create=True)
    scratchdir = os.path.join(tmpdir, 'scratch')
    os.mkdir(scratchdir)
    yield tmpdmf
    shutil.rmtree(tmpdir)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Workshop Modules.

Author: Jaffer Ghouse
"""
# stdlib
import os
from types import ModuleType
# third party
import pytest
# package
from idaes.util.testutil import run_notebook
from idaes.examples.workshops import (
    Module_1_Flash_Unit,
    Module_2_Flowsheet,
    Module_3_Custom_Unit_Model,
)


def module_path(modobj: ModuleType):
    return os.path.dirname(modobj.__file__)


# Test the 'solution' notebooks for all modules
# Note: modules 2 & 3 are skipped in circleci due to failure in free ipopt

def test_module_1():
    assert run_notebook(
        module_path(Module_1_Flash_Unit), "Module_1_Flash_Unit_Solution.ipynb"
    )


@pytest.mark.nocircleci()
def test_module_2():
    assert run_notebook(
        module_path(Module_2_Flowsheet), "Module_2_Flowsheet_Solution.ipynb"
    )


@pytest.mark.nocircleci()
def test_module_3():
    assert run_notebook(
        module_path(Module_3_Custom_Unit_Model), "Module_3_Exercise_1_Solution.ipynb"
    )



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.propindex module.
"""
# stdlib
import logging
import shutil
import sys
import tempfile

# third-party
import pytest

# package
import idaes
from idaes.dmf import propindex
from idaes.dmf import DMF
from idaes.dmf import resource

# for testing
from .util import init_logging
from . import for_propindex

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


def test_dmfvisitor():
    with pytest.raises(TypeError):
        propindex.DMFVisitor("o_O", "a")
    propindex.DMFVisitor("o_O")


@pytest.fixture
def tmpd():
    d = tempfile.mkdtemp(prefix="test_propindex_", suffix=".idaes")
    yield d
    shutil.rmtree(d)


@pytest.fixture
def testdmf(tmpd):
    dmf = DMF(path=tmpd, create=True)
    return dmf


def test_index_property_metadata(testdmf):
    propindex.index_property_metadata(
        testdmf, pkg=idaes.dmf, expr=".*IndexMePlease[0-9]", exclude_testdirs=False
    )
    # Check the resource
    for rsrc in testdmf.find():
        assert rsrc.v[rsrc.TYPE_FIELD] == resource.TY_CODE
        # print('@@ GOT RESOURCE:\n{}'.format(rsrc.v))


def test_index_multiple_versions(testdmf):
    v1, v2, v3 = "1.0.0", "6.6.6", "9.9.0"
    # index initial version
    propindex.index_property_metadata(
        testdmf,
        pkg=idaes.dmf,
        expr=".*IndexMePlease[0-9]",
        exclude_testdirs=False,
        default_version=v1,
    )
    # index again
    propindex.index_property_metadata(
        testdmf,
        pkg=idaes.dmf,
        expr=".*IndexMePlease[0-9]",
        exclude_testdirs=False,
        default_version=v2,
    )
    # check that we now have two resources, and
    # a relation between them
    rlist = list(testdmf.find({}))
    assert len(rlist) == 2
    rcodes = [r.v["codes"][0] for r in rlist]
    if rcodes[0]["version"][:3] == ("6", "6", "6"):
        first, second = 1, 0
    else:
        first, second = 0, 1

    # Debugging only
    # print('CODES:')
    # print(' - first -')
    # print(rcodes[first])
    # print(rlist[first].v['relations'])
    # print(' - second -')
    # print(rcodes[second])
    # print(rlist[second].v['relations'])

    # Each resource has 1 relation
    assert len(rlist[first].v["relations"]) == 1
    assert len(rlist[second].v["relations"]) == 1
    first_rel = rlist[first].v["relations"][0]
    second_rel = rlist[second].v["relations"][0]
    # First resource is pointed at by second
    assert first_rel[resource.RR_ROLE] == resource.RR_OBJ
    assert first_rel[resource.RR_PRED] == resource.PR_VERSION
    assert first_rel[resource.RR_ID] == rlist[second].id
    # Second resource points at first
    assert second_rel[resource.RR_ROLE] == resource.RR_SUBJ
    assert second_rel[resource.RR_PRED] == resource.PR_VERSION
    assert second_rel[resource.RR_ID] == rlist[first].id

    # Add the same version
    propindex.index_property_metadata(
        testdmf,
        pkg=idaes.dmf,
        expr=".*IndexMePlease[0-9]",
        exclude_testdirs=False,
        default_version=v2,
    )
    # check that we still have two resources
    rlist = list(testdmf.find({}))
    assert len(rlist) == 2

    # Now add another version
    propindex.index_property_metadata(
        testdmf,
        pkg=idaes.dmf,
        expr=".*IndexMePlease[0-9]",
        exclude_testdirs=False,
        default_version=v3,
    )
    # check that we now have three resources
    rlist = list(testdmf.find({}))
    assert len(rlist) == 3
    # check that we have 0 <--> 1 <--> 2
    # first sort by version and save that in the 'indexes' array
    indexes = [(r.v["codes"][0]["version"], i) for i, r in enumerate(rlist)]
    indexes.sort()
    # pull out relations into 'rel' array, in version order
    rel = [rlist[indexes[i][1]].v["relations"] for i in range(3)]
    # check first resource's relations
    assert len(rel[0]) == 1
    # 0 <-- 1
    assert rel[0][0][resource.RR_ID] == rlist[indexes[1][1]].id
    assert rel[0][0][resource.RR_ROLE] == resource.RR_OBJ
    # check second resource's relations
    assert len(rel[1]) == 2
    for j in range(2):
        if rel[1][j][resource.RR_ROLE] == resource.RR_SUBJ:
            # 1 --> 0
            assert rel[1][j][resource.RR_ID] == rlist[indexes[0][1]].id
        else:
            # 1 <-- 2
            assert rel[1][j][resource.RR_ID] == rlist[indexes[2][1]].id
            assert rel[1][j][resource.RR_ROLE] == resource.RR_OBJ
    # check third resource's relations
    assert len(rel[2]) == 1
    # 2 --> 1
    assert rel[2][0][resource.RR_ID] == rlist[indexes[1][1]].id
    assert rel[2][0][resource.RR_ROLE] == resource.RR_SUBJ




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Data Management Framework high-level functions.
"""
# stdlib
import logging
import sys
# package
from idaes.dmf import DMF
from idaes.dmf import errors
from idaes.dmf import propindex

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


def get_workspace(path='', name=None, desc=None, create=False, errs=None,
                  **kwargs):
    """Create or load a DMF workspace.

    If the :class:`DMF` constructor, throws an exception, this catches
    it and prints the error to the provided stream (or stdout).

    See :class:`DMF` for details on arguments.

    Args:
        path (str): Path to workspace.
        name (str): Name to be used for workspace.
        desc (str): Longer description of workspace.
        create (bool): If the path to the workspace does not exist,
                       this controls whether to create it.
        errs (object): Stream for errors, stdout is used if None

    Returns:
        DMF: New instance, or None if it failed.
    """
    dmf = None
    try:
        dmf = DMF(path=path, name=name, desc=desc, create=create, **kwargs)
    except errors.DMFError as err:
        if errs is None:
            errs = sys.stdout
        msg = 'Error creating DMF workspace\n'
        if isinstance(err, errors.DMFError) and not create:
            msg += 'Directory not found, and "create" flag is False\n'
            msg += 'If you want to create the workspace, try again with ' \
                   'create=True\n'
        else:
            msg += '{}\n'.format(err)
        msg += '\npath: {}\nname: {}\ndesc: {}\n'.format(path, name, desc)
        errs.write(msg)
    return dmf


def find_property_packages(dmf, properties=None):
    """Find all property packages matching provided criteria.

    Return the matching packages as a generator.

    Args:
        dmf (DMF): Data Management Framework instance
        properties (List[str]): Names of properties that must be present in the
              returned packages.
    Returns:
        Generator[idaes.dmf.resource.Resource]: Each object has the property
            data (properties and default units) in its `.data` attribute.
    """
    query = {'tags': [propindex.DMFVisitor.INDEXED_PROPERTY_TAG]}
    for prop in properties:
        field = 'data.properties.' + prop
        query[field] = True
    return dmf.find(query)


def index_property_packages(dmf):
    propindex.index_property_metadata(dmf)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Base class for unit models
"""

from pyomo.environ import Reference, SolverFactory
from pyomo.network import Port
from pyomo.opt import TerminationCondition
from pyomo.common.config import ConfigValue, In

from .process_base import (declare_process_block_class,
                           ProcessBlockData,
                           useDefault)
from .property_base import StateBlock
from .control_volume_base import ControlVolumeBlockData, FlowDirection
from idaes.core.util.exceptions import (BurntToast,
                                        ConfigurationError,
                                        PropertyPackageError)
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ['UnitModelBlockData', 'UnitModelBlock']

# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("UnitModelBlock")
class UnitModelBlockData(ProcessBlockData):
    """
    This is the class for process unit operations models. These are models that
    would generally appear in a process flowsheet or superstructure.
    """
    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("dynamic", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Dynamic model flag",
        doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}"""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Holdup construction flag",
        doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}"""))

    def build(self):
        """
        General build method for UnitModelBlockData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super(UnitModelBlockData, self).build()

        # Set up dynamic flag and time domain
        self._setup_dynamics()

    def model_check(blk):
        """
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume and tries to call the model_check method of the
        controlVolume block. If an AttributeError is raised, the check is
        passed.

        More complex models should overload this method with a model_check
        suited to the particular application, especially if there are multiple
        ControlVolume blocks present.

        Args:
            None

        Returns:
            None
        """
        # Run control volume block model checks
        try:
            blk.controlVolume.model_check()
        except AttributeError:
            pass

    def add_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build Port objects in a unit model and
        connect these to a specified StateBlock.
        Keyword Args:
            name = name to use for Port object.
            block = an instance of a StateBlock to use as the source to
                    populate the Port object
            doc = doc string for Port object
        Returns:
            A Pyomo Port object and associated components.
        """
        # Validate block object
        if not isinstance(block, StateBlock):
            raise ConfigurationError("{} block object provided to add_port "
                                     "method is not an instance of a "
                                     "StateBlock object. IDAES port objects "
                                     "should only be associated with "
                                     "StateBlocks.".format(blk.name))

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        p._state_block = (block, )

        # Get dict of Port members and names
        member_list = block[
                blk.flowsheet().config.time.first()].define_port_members()

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                slicer = block[:].component(member_list[s].local_name)
            else:
                slicer = block[:].component(member_list[s].local_name)[...]

            r = Reference(slicer)
            setattr(blk, "_"+name+"_"+s+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def add_inlet_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build inlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name = name to use for Port object (default = "inlet").
            block = an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the inlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc = doc string for Port object (default = "Inlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                        "{} add_inlet_port was called without a block argument"
                        " but a name argument was provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))
            else:
                name = "inlet"
            # Try for default ControlVolume name
            try:
                block = blk.control_volume
            except AttributeError:
                raise ConfigurationError(
                        "{} add_inlet_port was called without a block argument"
                        " but no default ControlVolume exists "
                        "(control_volume). Please provide block to which the "
                        "Port should be associated.".format(blk.name))
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                        "{} add_inlet_port was called with a block argument, "
                        "but a name argument was not provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))

        if doc is None:
            doc = "Inlet Port"

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        # Get dict of Port members and names
        if isinstance(block, ControlVolumeBlockData):
            try:
                member_list = (block.properties_in[
                                    block.flowsheet().config.time.first()]
                               .define_port_members())
                p._state_block = (block.properties_in, )
            except AttributeError:
                try:
                    member_list = (block.properties[
                                    block.flowsheet().config.time.first(), 0]
                                   .define_port_members())
                    if block._flow_direction == FlowDirection.forward:
                        p._state_block = (block.properties,
                                          block.length_domain.first())
                    elif block._flow_direction == FlowDirection.backward:
                        p._state_block = (block.properties,
                                          block.length_domain.last())
                except AttributeError:
                    raise PropertyPackageError(
                            "{} property package does not appear to have "
                            "implemented a define_port_memebers method. "
                            "Please contact the developer of the property "
                            "package.".format(blk.name))
        elif isinstance(block, StateBlock):
            member_list = block[
                    blk.flowsheet().config.time.first()].define_port_members()
            p._state_block = (block, )
        else:
            raise ConfigurationError(
                    "{} block provided to add_inlet_port "
                    "method was not an instance of a "
                    "ControlVolume or a StateBlock."
                    .format(blk.name))

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_in[:].component(
                                        member_list[s].local_name)
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.first()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.last()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx]
                                  .component(member_list[s].local_name))
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))
            else:
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_in[:].component(
                                        member_list[s].local_name)[...]
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.first()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.last()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx].component(
                                    member_list[s].local_name))[...]
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)[...]
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))

            r = Reference(slicer)
            setattr(blk, "_"+name+"_"+s+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def add_outlet_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build outlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name = name to use for Port object (default = "outlet").
            block = an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the outlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc = doc string for Port object (default = "Outlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                        "{} add_outlet_port was called without a block "
                        "argument  but a name argument was provided. Either "
                        "both a name and a block must be provided or neither."
                        .format(blk.name))
            else:
                name = "outlet"
            # Try for default ControlVolume name
            try:
                block = blk.control_volume
            except AttributeError:
                raise ConfigurationError(
                        "{} add_outlet_port was called without a block "
                        "argument but no default ControlVolume exists "
                        "(control_volume). Please provide block to which the "
                        "Port should be associated.".format(blk.name))
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                        "{} add_outlet_port was called with a block argument, "
                        "but a name argument was not provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))

        if doc is None:
            doc = "Outlet Port"

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        # Get dict of Port members and names
        if isinstance(block, ControlVolumeBlockData):
            try:
                member_list = (block.properties_out[
                                    block.flowsheet().config.time.first()]
                               .define_port_members())
                p._state_block = (block.properties_out, )
            except AttributeError:
                try:
                    member_list = (block.properties[
                                    block.flowsheet().config.time.first(), 0]
                                   .define_port_members())
                    if block._flow_direction == FlowDirection.forward:
                        p._state_block = (block.properties,
                                          block.length_domain.last())
                    elif block._flow_direction == FlowDirection.backward:
                        p._state_block = (block.properties,
                                          block.length_domain.first())
                except AttributeError:
                    raise PropertyPackageError(
                            "{} property package does not appear to have "
                            "implemented a define_port_members method. "
                            "Please contact the developer of the property "
                            "package.".format(blk.name))
        elif isinstance(block, StateBlock):
            member_list = block[
                    blk.flowsheet().config.time.first()].define_port_members()
            p._state_block = (block, )
        else:
            raise ConfigurationError(
                    "{} block provided to add_inlet_port "
                    "method was not an instance of a "
                    "ControlVolume or a StateBlock."
                    .format(blk.name))

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_out[:].component(
                                        member_list[s].local_name)
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.last()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.first()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx]
                                  .component(member_list[s].local_name))
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))
            else:
                # Need to use slice notation on indexed comenent as well
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_out[:].component(
                                        member_list[s].local_name)[...]
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.last()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.first()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx].component(
                                    member_list[s].local_name))[...]
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)[...]
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))

            r = Reference(slicer)
            setattr(blk, "_"+name+"_"+s+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe({"Inlet": self.inlet,
                                                  "Outlet": self.outlet},
                                                 time_point=time_point)
        except AttributeError:
            raise ConfigurationError(
                    f"Unit model {self.name} does not have the standard Port "
                    f"names (inet and outlet). Please contact the unit model "
                    f"developer to develop a unit specific stream table.")

    def initialize(blk, state_args=None, outlvl=6,
                   solver='ipopt', optarg={'tol': 1e-6}):
        '''
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume, and first initializes this and then attempts to solve
        the entire unit.

        More complex models should overload this method with their own
        initialization routines,

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = {}).
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating which solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        '''
        # Set solver options
        init_log = getInitLogger(blk.name, outlvl)
        opt = SolverFactory(solver)
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Initialize control volume block
        flags = blk.control_volume.initialize(outlvl=outlvl+1,
                                              optarg=optarg,
                                              solver=solver,
                                              state_args=state_args)

        init_log.log(4, 'Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        # Solve unit
        results = opt.solve(blk, tee=init_tee(init_log))

        init_log.log(4, "Initialization Step 2 {}.".format(condition(results)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.control_volume.release_state(flags, outlvl+1)

        init_log.log(5, 'Initialization Complete: {}'.format(condition(results)))
